# Copyright (c) 2025 Itential, Inc
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from typing import Annotated

from pydantic import Field

from fastmcp import Context

from itential_mcp import functions


async def get_jobs(
    ctx: Annotated[Context, Field(
        description="The FastMCP Context object"
    )],
    name: Annotated[str | None, Field(
        description="Workflow name used to filter the results",
        default=None
    )],
    project: Annotated[str | None, Field(
        description="Project name used to filter the results",
        default=None
    )]
) -> list[dict]:
    """
    Get all jobs from the Itential Platform server

    This tool will retrieve all jobs from the Itential Platform server and
    return the metdata from the jobs as a list.  The metadata includes the
    job id and name, current job status, and the job description if it was
    set.

    This tool has two optional arguments that can be uesd to filter the
    list of returned jobs.  The first optional argument is name which
    defines the name of the workflow to return jobs for.  This will restrict
    the returned list to only workflows that match this name.

    The second optional argument is project.  By default, the returned list
    only considers workflows in the global namespace.  If the workfow is
    embedded in a project, the project argument must be provided.  This
    argument accepts the project name that contains the workflow.

    If the project argument is specified and the name argument is not
    specified, all workflow jobs assoicated with the project are
    returned.

    The following data is returned for each job in the list:

        * _id: The unique job identifier
        * created: The timestamp when the job was created
        * createdBy: The id of the user that created the job
        * description: A short description of the job created by the user
        * updated: The timestamp of when the job was last updated
        * updatedBy: The id of user that last updated the job
        * name: The name of the job
        * status: The current status of the job.  The job status will be one
            of `error`, `complete`, `running`, `cancelled`, `incomplete`, or
            `paused`

    Args:
        ctx (Context): The FastMCP Context object

        name (str): Only return jobs for workflows that match this value

    Returns:
        list[dict]: A list of Python dict objects where each element
            represents the metadata for a single job

    Raises:
        None:
    """
    await ctx.info("running get_jobs(...)")

    client = ctx.request_context.lifespan_context.get("client")

    results = list()

    limit = 100
    skip = 0

    params = {"limit": limit}

    if project is not None:
        project_id = await functions.project_name_to_id(ctx, project)
        if name is not None:
            params["equals[name]"] = f"@{project_id}: {name}"
        else:
            params["starts-with[name]"] = f"@{project_id}"

    elif name is not None:
        params["equals[name]"] = name

    while True:
        params["skip"] = skip

        res = await client.get("/operations-manager/jobs", params=params)

        data = res.json()
        metadata = data.get("metadata")

        for item in data.get("data") or list():
            results.append({
                "_id": item.get("_id"),
                "created": item.get("created"),
                "created_by": item.get("created_by"),
                "updated": item.get("last_updated"),
                "updated_by": item.get("last_updated_by"),
                "name": item.get("name"),
                "description": item.get("description"),
                "status": item.get("status")
            })

        if len(results) == metadata["total"]:
            break

        skip += limit

    return results


async def describe_job(
    ctx: Annotated[Context, Field(
        description="The FastMCP Context object"
    )],
    job_id: Annotated[str, Field(
        description="The ID used to retreive the job"
    )]
) -> dict:
    """
    Get details about a job from Itential Platform

    When a workflow is started a new job is automatically created that
    contains the status of the job.  All details about the job are stored
    in the job document.  The job document provides information about
    the execution of a workflow.

    This function will retrieve the job document from Itential Platform based
    on the unique `job_id` argument.   The `job_id` is used to uniquely
    identify the desired job document.

    The job document will return the following:
        * _id: The unique identifier for this job
        * name: The name of the API endpoint trigger
        * description: Short description of the API endpoint trigger
        * type: Identifies the type of job.  Valid values for type are
            `automation`, `resource:action`, or `resource:compliance`
        * tasks: The full set of tasks to be executed
        * updated: ISO 8601 timestamp of when the trigger was last updated
        * status: The status of the job.  This will return one of the
            following values: `error`, `complete`, `running`, `canceled`,
            `incomplete` or `paused`
        * metrics: Job metrics that include the job start time, job end
            time and account

    Args:
        ctx (Context): The FastMCP Context object

        job_id (str): The job identifier returned from the job _id returned
            for any triggered job

    Returns:
        dict: A Python dict object that represents the job status from the
            server

    Raises:
        None
    """

    await ctx.info("inside describe_job(...)")

    client = ctx.request_context.lifespan_context.get("client")

    res = await client.get(f"/operations-manager/jobs/{job_id}")

    data = res.json()["data"]

    return {
        "_id": data["_id"],
        "name": data["name"],
        "description": data["description"],
        "type": data["type"],
        "tasks": data["tasks"],
        "status": data["status"],
        "metrics": data["metrics"],
        "updated": data["last_updated"]
    }



# Copyright (c) 2025 Itential, Inc
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

import os
import configparser

from functools import lru_cache
from pathlib import Path

from typing import Literal
from dataclasses import fields

from pydantic import Field
from pydantic.dataclasses import dataclass

from . import env


def options(*args, **kwargs) -> dict:
    """
    Utility function to add extra parameters to fields

    This function will add extra parameters to to a Field in the Config
    class.  Specifically it handles adding the necessary keys to support
    generating the CLI options from the configuration.  This unifies the
    parameter descriptions and default values for consistency.

    Args:
        *args: Positional arguments to be added to the CLI command line option
        **kwargs: Optional arguments to be added to the CLI command line option

    Returns:
        dict: A Python dict object to be added to the Field function
            signature

    Raises:
        None
    """
    return {
        "x-itential-mcp-cli-enabled": True,
        "x-itential-mcp-arguments": args,
        "x-itential-mcp-options": kwargs
    }

@dataclass(frozen=True)
class Config(object):

    server_transport: Literal["stdio", "sse", "streamable-http"] = Field(
        description="The MCP server transport to use",
        default="stdio",
        json_schema_extra=options(
            "--transport",
            choices=("stdio", "sse", "streamable-http"),
            metavar="<value>"
        )
    )


    server_host: str = Field(
        description="Address to listen for connections on",
        default="127.0.0.1",
        json_schema_extra=options(
            "--host",
            metavar="<host>"
        )
    )

    server_port: int = Field(
        description="Port to listen for connections on",
        default=8000,
        json_schema_extra=options(
            "--port",
            metavar="<port>",
            type=int
        )
    )

    server_path: str = Field(
        description="URI path used to accept requests from",
        default="/mcp",
        json_schema_extra=options(
            "--path",
            metavar="<path>"
        )
    )

    server_log_level: Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"] = Field(
        description="Logging level for verbose output",
        default="INFO",
        json_schema_extra=options(
            "--log-level",
            metavar="<level>"
        )
    )

    server_include_tags: str | None = Field(
        description="Include tools that match at least on tag",
        default=None,
        json_schema_extra=options(
            "--include-tags",
            metavar="<tags>"
        )
    )

    server_exclude_tags: str | None = Field(
        description="Exclude any tool that matches one of these tags",
        default="experimental,beta",
        json_schema_extra=options(
            "--exclude-tags",
            metavar="<tags>"
        )
    )

    platform_host: str = Field(
        description="The host addres of the Itential Platform server",
        default="localhost",
        json_schema_extra=options(
            "--platform-host",
            metavar="<host>"
        )
    )

    platform_port: int = Field(
        description="The port to use when connecting to Itential Platform",
        default=0,
        json_schema_extra=options(
            "--platform-port",
            type=int,
            metavar="<port",
        )
    )

    platform_disable_tls: bool = Field(
        description="Disable using TLS to connect to the server",
        default=False,
        json_schema_extra=options(
            "--platform-disable-tls",
            action="store_true"
        )
    )

    platform_disable_verify: bool = Field(
        description="Disable certificate verification",
        default=False,
        json_schema_extra=options(
            "--platform-disable-verify",
            action="store_true"
        )
    )

    platform_user: str = Field(
        description="Username to use when authenticating to the server",
        default="admin",
        json_schema_extra=options(
            "--platform-user",
            metavar="<user>"
        )
    )

    platform_password: str = Field(
        description="Password to use when authenticating to the server",
        default="admin",
        json_schema_extra=options(
            "--platform-password",
            metavar="<password>"
        )
    )

    platform_client_id: str | None = Field(
        description="Client ID to use when authenticating using OAuth",
        default=None,
        json_schema_extra=options(
            "--platform-client-id",
            metavar="<client_id>"
        )
    )

    platform_client_secret: str | None = Field(
        description="Client secret to use when authenticating using OAuth",
        default=None,
        json_schema_extra=options(
            "--platform-client-secret",
            metavar="<client_secret>"

        )
    )

    platform_timeout: int = Field(
        description="Sets the timeout in seconds when communciating with the server",
        default=30,
        json_schema_extra=options(
            "--platform-timeout",
            metavar="<secs>"
        )
    )

    @property
    def server(self):
        return {
            "transport": self.server_transport,
            "host": self.server_host,
            "port": self.server_port,
            "path": self.server_path,
            "log_level": self.server_log_level,
            "include_tags": self._coerce_to_set(self.server_include_tags) if self.server_include_tags else None,
            "exclude_tags": self._coerce_to_set(self.server_exclude_tags) if self.server_exclude_tags else None
        }

    @property
    def platform(self):
        return {
            "host": self.platform_host,
            "port": self.platform_port,
            "use_tls": not self.platform_disable_tls,
            "verify": not self.platform_disable_verify,
            "user": self.platform_user,
            "password": self.platform_password,
            "client_id": None if self.platform_client_id == "" else self.platform_client_id,
            "client_secret": None if self.platform_client_secret == "" else self.platform_client_secret,
            "timeout": self.platform_timeout
        }

    def _coerce_to_set(self, value) -> list:
        items = set()
        for ele in value.split(","):
            items.add(ele.strip())
        return items


@lru_cache(maxsize=None)
def get() -> Config:
    """
    Return the configuration instance

    This function will load the configuration and return an instance of
    Config.  This function is cached and is safe to call multiple times.
    The configuration is loaded only once and the cached Config instance
    is returned with every call.

    Args:
        None

    Returns:
        Conig: An instance of Config that represents the application
            configuration

    Raises:
        FileNotFoundError: If a configuration file is specified but not found
            this exception is raised
    """
    conf_file = env.getstr("ITENTIAL_MCP_CONFIG")

    data = {}

    if conf_file is not None:
        path = Path(conf_file)
        if not path.is_file():
            raise FileNotFoundError(f"Config file not found: {path}")

        cf = configparser.ConfigParser()
        cf.read(conf_file)

        for item in cf.sections():
            for key, value in cf.items(item):
                key = f"{item}_{key}"
                data[key] = value

    for item in fields(Config):
        envkey = f"ITENTIAL_MCP_{item.name}".upper()
        if envkey in os.environ:
            value = ", ".join(os.environ[envkey].split(","))
            data[item.name] = value

    return Config(**data)



# Copyright (c) 2025 Itential, Inc
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

import asyncio

from .cli import run

if __name__ == "__main__":
    asyncio.run(run())



# Copyright (c) 2025 Itential, Inc
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

import sys
import asyncio

from unittest.mock import patch, MagicMock, AsyncMock

import pytest

from itential_mcp import server


@pytest.mark.asyncio
async def test_lifespan_yields_client_and_cache():
    mcp = MagicMock()
    async with server.lifespan(mcp) as context:
        assert "client" in context
        assert "cache" in context
        assert context["client"].__class__.__name__ == "PlatformClient"
        assert context["cache"].__class__.__name__ == "Cache"


def test_register_tools(tmp_path, monkeypatch):
    # Setup a temporary tools directory with a sample tool module
    tools_dir = tmp_path / "tools"
    tools_dir.mkdir()
    sample_tool = tools_dir / "echo.py"
    sample_tool.write_text(
        "def hello():\n"
        "    return 'world'\n"
    )

    # Patch __file__ to point to fake module with tools path
    monkeypatch.setattr(server, "__file__", str(tmp_path / "fake.py"))

    mcp = MagicMock()
    server.register_tools(mcp)

    # Tool `hello` from `echo` module should be added
    assert mcp.tool.called
    args, kwargs = mcp.tool.call_args
    assert args[0].__name__ == "hello"
    assert "tags" in kwargs
    assert "hello" in kwargs["tags"]


@pytest.mark.asyncio
@patch("itential_mcp.server.register_tools")
@patch("itential_mcp.server.FastMCP.run_async", new_callable=AsyncMock)
@patch("itential_mcp.server.config.get")
async def test_run_stdio_success(mock_config_get, mock_run_async, mock_register_tools):
    mock_config = MagicMock()
    mock_config.server = {
        "transport": "stdio",
        "log_level": "INFO",
        "include_tags": [],
        "exclude_tags": [],
    }
    mock_config_get.return_value = mock_config

    result = await server.run()
    assert mock_run_async.called
    assert result is None or result == 0


@patch("itential_mcp.server.register_tools", side_effect=Exception("tool error"))
@patch("itential_mcp.server.config.get")
def test_run_tool_import_error(mock_config_get, mock_register_tools):
    mock_config = MagicMock()
    mock_config.server = {
        "transport": "stdio",
        "log_level": "INFO",
        "include_tags": [],
        "exclude_tags": [],
    }
    mock_config_get.return_value = mock_config

    with patch.object(sys, "stderr"), pytest.raises(SystemExit) as excinfo:
        asyncio.run(server.run())

    assert excinfo.value.code == 1




# Copyright (c) 2025 Itential, Inc
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

import os
import configparser

import pytest

from itential_mcp import config as config_module


@pytest.fixture(autouse=True)
def clear_config_cache():
    """Ensure config.get() doesn't cache between tests"""
    config_module.get.cache_clear()
    yield
    config_module.get.cache_clear()


def test_get_config_from_env(monkeypatch):
    monkeypatch.setenv("ITENTIAL_MCP_SERVER_HOST", "127.0.0.1")
    monkeypatch.setenv("ITENTIAL_MCP_SERVER_PORT", "1234")
    monkeypatch.setenv("ITENTIAL_MCP_PLATFORM_USER", "testuser")
    monkeypatch.setenv("ITENTIAL_MCP_PLATFORM_PASSWORD", "secret")
    monkeypatch.setenv("ITENTIAL_MCP_PLATFORM_DISABLE_TLS", "true")

    cfg = config_module.get()

    assert cfg.server_host == "127.0.0.1"
    assert cfg.server_port == 1234
    assert cfg.platform_user == "testuser"
    assert cfg.platform_password == "secret"
    assert cfg.platform_disable_tls is True


def test_get_config_from_file(tmp_path, monkeypatch):
    config_path = tmp_path / "test.ini"

    cp = configparser.ConfigParser()
    cp["server"] = {
        "host": "192.168.1.1",
        "port": "9000"
    }
    cp["platform"] = {
        "user": "fileuser",
        "password": "filepass",
        "disable_tls": "true"
    }

    with open(config_path, "w") as f:
        cp.write(f)

    for ele in os.environ.keys():
        if ele.startswith("ITENTIAL"):
            monkeypatch.delenv(ele, raising=False)

    monkeypatch.setenv("ITENTIAL_MCP_CONFIG", str(config_path))

    cfg = config_module.get()

    assert cfg.server_host == "192.168.1.1"
    assert cfg.server_port == 9000
    assert cfg.platform_user == "fileuser"
    assert cfg.platform_password == "filepass"
    assert cfg.platform_disable_tls is True


def test_missing_config_file_raises(monkeypatch):
    monkeypatch.setenv("ITENTIAL_MCP_CONFIG", "/nonexistent/path.ini")

    with pytest.raises(FileNotFoundError):
        config_module.get()


def test_config_platform_and_server_properties(monkeypatch):
    monkeypatch.setenv("ITENTIAL_MCP_SERVER_INCLUDE_TAGS", "public,system")
    monkeypatch.setenv("ITENTIAL_MCP_SERVER_EXCLUDE_TAGS", "experimental,beta")

    cfg = config_module.get()

    assert cfg.server["include_tags"] == {"public", "system"}
    assert cfg.server["exclude_tags"] == {"experimental", "beta"}
    assert isinstance(cfg.platform, dict)
    assert "host" in cfg.platform




# Copyright (c) 2025 Itential, Inc
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

import pytest
import time

from itential_mcp.cache import Cache


@pytest.fixture
def cache():
    c = Cache(cleanup_interval=1)
    yield c
    c.stop()


def test_put_get(cache):
    cache.put("foo", "bar")
    assert cache.get("foo") == "bar"


def test_get_nonexistent_key(cache):
    assert cache.get("missing") is None


def test_put_with_ttl(cache):
    cache.put("foo", "bar", ttl=1)
    time.sleep(1.5)
    assert cache.get("foo") is None


def test_delete_key(cache):
    cache.put("foo", "bar")
    cache.delete("foo")
    assert cache.get("foo") is None


def test_delete_nonexistent_key(cache):
    # Should not raise any exceptions
    assert cache.delete("ghost") is None


def test_keys_include_and_expired(cache):
    cache.put("foo", "bar", ttl=2)
    cache.put("baz", "qux")
    time.sleep(1)
    assert "foo" in cache.keys()
    assert "baz" in cache.keys()
    time.sleep(2)
    assert "foo" not in cache.keys()
    assert "baz" in cache.keys()


def test_clear(cache):
    cache.put("a", 1)
    cache.put("b", 2)
    cache.clear()
    assert cache.get("a") is None
    assert cache.get("b") is None
    assert cache.keys() == []


def test_stop_clears_cache():
    c = Cache(cleanup_interval=1)
    c.put("x", "y")
    c.stop()
    assert c.get("x") is None




# Copyright (c) 2025 Itential, Inc
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from typing import Annotated

from pydantic import Field

from fastmcp import Context


async def get_device_groups(
    ctx: Annotated[Context, Field(
        description="The FastMCP Context object"
    )],
) -> list[dict]:
    """
    Get all device groups from Itential Platform server

    This tool will retrieve all of the configured device groups from
    Itential Platform server.  The response includes a list of elements
    where each element represents a unique device group.

    Each element has the following fields:

        * id: The unique identifier for the group
        * name: The name of the device group
        * devices: The list of device names that comprise this group
        * description: A short description of this device group
        * created: ISO 8601 timestamp of when the trigger was created
        * createdBy: Account name that created the trigger
        * updated: ISO 8601 timestamp of when the trigger was last updated
        * updatedBy: Account name that last updated the trigger
        * gbac: Returns the groups that have read and write access to this
            device group

    Args:
        ctx (Context): The FastMCP Context object

    Returns:
        dict: A Python list of dict objects that represents the list
            of device groups configured on Itential Platform

    Raises:
        None
    """
    await ctx.info("inside get_device_groups(...)")

    client = ctx.request_context.lifespan_context.get("client")

    results = list()

    res = await client.get("/configuration_manager/deviceGroups")

    data = res.json()

    for ele in data:
        results.append({
            "id": ele["id"],
            "name": ele["name"],
            "devices": ele["devices"],
            "description": ele["description"],
        })

    #for ele in data:
    #    for gbac in ("read", "write"):
    #        items = list()
    #        for item in ele["gbac"][gbac]:
    #            items.append(await functions.group_id_to_name(ctx, item))
    #        ele["gbac"][gbac] = items
    #    results.append(ele)

    return results


async def create_device_group(
    ctx: Annotated[Context, Field(
        description="The FastMCP Context object"
    )],
    name: Annotated[str, Field(
        description="The name of the device group to create"
    )],
    description: Annotated[str | None, Field(
        description="Short description of the device group",
        default=None
    )],
    devices: Annotated[list | None, Field(
        description="List of devices to add to the group",
        default=None
    )]
) -> dict:
    """
    Create a new device group on Itential Server

    This tool will create a new device group on the server.  It has one
    required argument `name` which defines the name of the group to
    create.  If a group with the same name already exists, an
    error is returned, otherwise the group is created.

    The optional `description` argument sets a short description of
    the device group.

    The optional `devices` argument configures the list of devices to
    include in the device group.  The list of devices should be devices
    known to Itential Platform.  The list of devices can be found
    using the `get_devices` tool.

    This tool will return a message to indicate the device was created
    successfully.

        * id: The unqiue identifer for this device group
        * name: The name of the device group
        * message: Short status message describing the create operation
        * status: Current status of the device group.

    Args:
        ctx (Context): The FastMCP Context object

        name (str): The name of the device group to create

        description (str): Short description of the device group

        devices: (list): List of devices to add to the group when it is
            created

    Returns:
        dict: An object that provides the status of the device group
            create operation

    Raises:
        ValueError: Raised if a device group by the same name already
            exists on the server
    """
    await ctx.info("inside create_device_group(...)")

    client = ctx.request_context.lifespan_context.get("client")

    groups = await get_device_groups(ctx)

    for ele in groups:
        if ele["name"] == name:
            raise ValueError(f"device group {name} already exists")

    body = {
        "groupName": name,
        "groupDescription": description
    }

    if devices:
        body["deviceNames"] = ",".join(devices)

    res = await client.post(
        "/configuration_manager/devicegroup",
        json=body
    )

    return res.json()



# Copyright (c) 2025 Itential, Inc
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

class ItentialMcpError(Exception):
    """
    Base exception class for all Itential MCP errors

    This is the base exception class that all errors generated by the
    Itential MCP server should derive from.
    """
    pass

class TimeoutExceededError(ItentialMcpError):
    """
    Error raised when a timeout value has been exceeded.

    This exception is typically used by tools to indicate an operation on
    the server has taken longer than a configured timeout value.
    """
    pass


class NotFoundError(ItentialMcpError):
    """
    Error raised when a specific resource could not be found

    This exception is typically used by tools to indicate that a resource
    could not be found on the server.
    """
    pass


class InvalidStateError(ItentialMcpError):
    """
    Error raised when a resource is in an unexpected or invalid state

    This exception is typically used by tools to indicate that a resource is
    in a state that was either unexpected or invalid.
    """
    pass



# Copyright (c) 2025 Itential, Inc
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from typing import Callable


def tags(*tag_list) -> Callable:
    """
    Decorator that will add tags to a function

    This decorator when called will add one or more tags to a function.  The
    tags are used to control which tools are exposed when the server is
    started.

    To use this decoator, import the function into the tools module and
    decorate the target tool as shown below.

    ```
    from itential_mcp.toolutils import tags

    @tags("public", "system")
    def get_server_info(ctx: Context) -> dict:
        return {}
    ```

    Args:
        *tag_list: The list of tags to be attached to the function

    Returns:
        Callable: A callable decorated function

    Raises:
        None
    """
    def decorator(func):
        setattr(func, "tags", list(tag_list))
        return func
    return decorator



# Copyright (c) 2025 Itential, Inc
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from typing import Annotated

from pydantic import Field

from fastmcp import Context

from itential_mcp import timeutils
from itential_mcp import functions


async def get_workflows(
    ctx: Annotated[Context, Field(
        description="The FastMCP Context object"
    )]
) -> list[dict]:
    """
    Get all API endpoint triggers from Itential Platform

    This tool will retrieve all of the API endpoint triggers (routes) that
    can be called by external services.  The tool will return a list of
    triggers where each element in the list represents a callable API
    route. Use the name as the identifier for the workflow. The routeName 
    is used for job triggering only.

    The fields available for each element include:

        * _id: The unique identifier for this route
        * name: The name of the workflow as configured in Platform
        * description: Short description of the triggers function
        * schema: The input schema for the launching the trigger based
            on http://json-schema.org/draft-07/schema
        * routeName: The API route used to start the endpoint
        * created: ISO 8601 timestamp of when the trigger was created
        * createdBy: Account name that created the trigger
        * updated: ISO 8601 timestamp of when the trigger was last updated
        * updatedBy: Account name that last updated the trigger
        * lastExecuted: ISO 8601 timestamp this trigger was last executed

    Args:
        ctx (Context): The FastMCP Context object

    Returns:
        list[dict]: A Python list of dict objects that represent the available
            workflows found on the server.

    Raises:
        None
    """
    await ctx.info("inside get_workflows(...)")

    client = ctx.request_context.lifespan_context.get("client")

    limit = 100
    skip = 0

    params = {"limit": limit}

    results = list()

    while True:
        params.update({
            "skip": skip,
            "equalsField": "type",
            "equals": "endpoint",
            "enabled": True,
        })

        res = await client.get(
            "/operations-manager/triggers",
            params=params,
        )

        data = res.json()

        for item in data.get("data") or list():

            if item.get("lastExecuted") is not None:
                lastExecuted = timeutils.epoch_to_timestamp(item["lastExecuted"])
            else:
                lastExecuted = None

            results.append({
                "_id": item.get("_id"),
                "name": item.get("name"),
                "description": item.get("description"),
                "schema": item.get("schema"),
                "routeName": item.get("routeName"),
                "created": item.get("created"),
                "createdBy": item.get("createdBy"),
                "updated": item.get("lastUpdated"),
                "updatedBy": item.get("lastUpdatedBy"),
                "lastExecuted": lastExecuted,
            })

        if len(results) == data["metadata"]["total"]:
            break

        skip += limit

    return results


async def start_workflow(
    ctx: Annotated[Context, Field(
        description="The FastMCP Context object"
    )],
    route_name: Annotated[str, Field(
        description="The name of the API endpoint used to start the workflow"
    )],
    data: Annotated[dict | None, Field(
        description="Data to include in the request body when calling the route",
        default=None
    )]
) -> dict:
    """
    Start an API endpoint trigger from Itential Platform

    This tool is responsible for calling the API endpoint trigger based
    on the `route` argument.  The `route` argument provides the name of
    API endpoint trigger to call.

    Optionally, the `data` argument can be used to provide data in the
    body of the request when invoking the API endpoint route.

    The tool will return an object that represents the started job
    in Itential Platform.  It containes the following fields:

        * _id: The unique identifier for this job
        * name: The name of the API endpoint trigger
        * description: Short description of the API endpoint trigger
        * tasks: The full set of tasks to be executed
        * created: ISO 8601 timestamp of when the trigger was created
        * createdBy: Account name that created the trigger
        * updated: ISO 8601 timestamp of when the trigger was last updated
        * updatedBy: Account name that last updated the trigger
        * status: The status of the job.  This will return one of the
            following values: `error`, `complete`, `running`, `canceled`,
            `incomplete` or `paused`
        * metrics: Job metrics that include the job start time, job end
            time and account

    Args:
        ctx (Context): The FastMCP Context object
        route_name (str): The API endpoint route to start
        data: (dict): Data to use in the request as input to the API route

    Returns:
        dict: A Python dict object that represents the relevant data from
            the job document API.

    Raises:
        None

    """
    await ctx.info("inside run_workflow(...)")

    client = ctx.request_context.lifespan_context.get("client")

    res = await client.post(
        f"/operations-manager/triggers/endpoint/{route_name}",
        json=data,
    )

    data = res.json()["data"]

    metrics = {}

    metrics_data = data.get("metrics") or {}

    if metrics_data.get("start_time") is not None:
        metrics["start_time"] = timeutils.epoch_to_timestamp(metrics_data["start_time"])

    if metrics_data.get("end_time") is not None:
        metrics["end_time"] = timeutils.epoch_to_timestamp(metrics_data["end_time"])

    if metrics_data.get("user") is not None:
        metrics["user"] = await functions.account_id_to_username(ctx, metrics_data["user"])

    return {
        "_id": data["_id"],
        "name": data["name"],
        "description": data["description"],
        "tasks": data["tasks"],
        "status": data["status"],
        "metrics": metrics,
        "updated": data["last_updated"],
        "updated_by": data["last_updated_by"],
        "created": data["created"],
        "created_by": data["created_by"],
    }



# Copyright (c) 2025 Itential, Inc
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from typing import Annotated

from pydantic import Field

from fastmcp import Context


async def get_health(
    ctx: Annotated[Context, Field(
        description="The FastMCP Context object"
    )]
) -> dict:
    """
    Get the Itential Platforms server health

    The server health response will return data about the overall health of
    the server. It will return keys that include the current server status
    including details about memory and CPU utilization, version information
    for the server and dependencies as well as the system CPU architecture.

    The health response includes keys for `status` which provides the overall
    system status including core dependent service status for mongo and redis.
    It also includes a key for `system` the provides the server architecture,
    total memory and CPU core details.  The `server` key provides details about
    the server software including running versions, memory and CPU usage details
    and dependent library versions.

    The health response also includes details about running applications and
    adapters. The `applications` key provides details about applications and
    the `adapters` key provides detailsa about adapters. This includes the
    current running state of the application or adatper as well as the
    memory and CPU usage statistics and uptime.

    Args:
        ctx (Context): The FastMCP Context object

    Returns:
        dict: A Python dict object with the response

    Raises:
        None
    """
    await ctx.info("inside get_health(...)")

    client = ctx.request_context.lifespan_context.get("client")

    results = {}

    for key, uri in (
        ("status", "/health/status"),
        ("system", "/health/system"),
        ("server", "/health/server"),
        ("applications", "/health/applications"),
        ("adapters", "/health/adapters"),
    ):
        res = await client.get(uri)
        data = res.json()
        results[key] = data.get("results") or data

    return results



# Copyright (c) 2025 Itential, Inc
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

import time

from typing import Annotated

from pydantic import Field

from fastmcp import Context

from itential_mcp import errors


async def _get_application_health(
    ctx: Context,
    name: str
) -> dict:
    """
    Get the health of an application

    This internal function will return the the health of the application
    specified by the name argument.  If the specified application does not
    exist, the function will raise a NotFoundError.   If the application
    does exist, it will be returned as a dict object.

    Args:
        ctx (Context): The FastMCP Context object

        name (str): The name of the application to get.  This argument is
            case senstive.  To get a list of all available applications on
            the server see ```get_applications```

    Returns:
        dict: An object the represents the application health

    Raises:
        NotFoundError: If the application specified by name cannot be found
            on the server, a NotFoundError is returned.
    """
    client = ctx.request_context.lifespan_context.get("client")

    res = await client.get(
        "/health/applications",
        params={
            "equals": name,
            "equalsField": "id"
        }
    )

    data = res.json()

    if data["total"] != 1:
        raise errors.NotFoundError(f"unable to find application {name}")

    return data


async def get_applications(
    ctx: Annotated[Context, Field(
        description="The FastMCP Context object"
    )]
) -> dict:
    """
    Get all applications from Itential Platform

    This tool will retrieve all of the applications configured on the instance
    of Itential Plaform and return them.   The returned list will include
    the application name, package, version, description and running state.

    Args:
        ctx (Context): The FastMCP Context object

    Returns:
        list[dict]: A list of objects where each object represents an
            application with the following fields:
            - name: The name of the application
            - package: The NodeJS package that comprises the application
            - version: The version of the application on the server
            - description: The application description
            - state: The current operational state of the appliacation.  Valid
                states are DEAD, STOPPED, RUNNING, DELETED

    Raises:
        None
    """
    await ctx.info("inside get_applications(...)")

    client = ctx.request_context.lifespan_context.get("client")

    res = await client.get("/health/applications")

    data = res.json()

    results = list()

    for ele in data["results"]:
        results.append({
            "name": ele["id"],
            "package": ele.get("package_id"),
            "version": ele["version"],
            "description": ele.get("description"),
            "state": ele["state"],
        })

    return results


async def start_application(
    ctx: Annotated[Context, Field(
        description="The FastMCP Context object"
    )],
    name: Annotated[str, Field(
        description="The name of the application to start"
    )],
    timeout: Annotated[int, Field(
        description="Timeout waiting for application to start",
        default=10
    )]
) -> dict:
    """
    Start an application on Itential Platform

    This tool will start an application on an Itential Platform server.  The
    name argument defines the name of the application to start.  If the
    application is already in a running state. this tool will not attempt
    any further action.

    The name argument is case sensitive and must be a valid application
    defined on Itential Platform.  The list of applications running on
    the server can be found using the get_applications tool.

    If the application is in a STOPPED state, this tool will attempt to start
    the application.   It will check to see if the application started.  If
    the application hasn't started within the timeout period, a
    TimeoutExceededError error is raised

    If the application is in either the DEAD or DELETED state, an
    InvalidStateError is raised.

    Args:
        ctx (Context): The FastMCP Context object

        name (str): The name of the application to start

        timeout (int): The timeout in seconds this tool should wait for the
            application to start successfully.  If the application doesn't
            start before the timeout expires, a TimeoutExceeded exception
            is raised

    Returns:
        dict: An object representing the start operation
            - name: The name of the application
            - status: The state of the start operation.  Valid values are
                RUNNING, DEAD, DELETED, STOPPED

    Raises:
        TimeoutExceededError: If the application isn't in a RUNNING state
            before the timeout expires

        InvalidStateError: if the application is in either the DEAD or
            DELETED state, this exception is raised

    Notes:
        - The application name argument is case senstive
        - The timeout value is set in seconds

    """
    await ctx.info("inside start_application(...)")

    client = ctx.request_context.lifespan_context.get("client")

    data = await _get_application_health(ctx, name)
    state = data["results"][0]["state"]

    if state == "STOPPED":
        await client.put(f"/applications/{name}/start")

        while timeout:
            data = await _get_application_health(ctx, name)
            state = data["results"][0]["state"]

            if state == "RUNNING":
                break

            time.sleep(1)
            timeout -= 1

    elif state in ("DEAD", "DELETED"):
        raise errors.InvalidStateError(f"application `{name}` is `{state}`")

    if timeout == 0:
        raise errors.TimeoutExceededError()

    return {
        "name": name,
        "state": state
    }


async def stop_application(
    ctx: Annotated[Context, Field(
        description="The FastMCP Context object"
    )],
    name: Annotated[str, Field(
        description="The name of the application to stop"
    )],
    timeout: Annotated[int, Field(
        description="Timeout waiting for application to start",
        default=10
    )]
) -> dict:
    """
    Stop an application on Itential Platform

    This tool will stop an application that is currently in a running
    state on the server.  If the application is already stopped this
    tool will not perform any further action.

    The name argument specifies the name of the application to stop.  The
    value for name must be a valid case senstive name.  The list of
    applications can be found by calling the get_applications tool.  If
    the specified name is not a valid application name, this tool will
    raise a NotFoundError.

    The timeout value sets the number of seconds to wait to verify the
    application has stopped.  If the timeout value expires before the
    application reaches a stopped state, this tool will raise a
    TimeoutExceededError.

    If the application is in either the DEAD or DELETED state, an
    InvalidStateError is raised.

    Args:
        ctx (Context): The FastMCP Context object

        name (str): The name of the application to start

        timeout (int): The timeout in seconds this tool should wait for the
            application to stop successfully.  If the application doesn't
            stop before the timeout expires, a TimeoutExceeded exception
            is raised

    Returns:
        dict: An object representing the stop operation
            - name: The name of the application
            - status: The state of the start operation.  Valid values are
                RUNNING, DEAD, DELETED, STOPPED

    Raises:
        TimeoutExceededError: If the application isn't in a STOPPED state
            before the timeout expires

        InvalidStateError: if the application is in either the DEAD or
            DELETED state, this exception is raised

    Notes:
        - The application name argument is case senstive
        - The timeout value is set in seconds
    """
    await ctx.info("inside stop_application(...)")

    client = ctx.request_context.lifespan_context.get("client")

    data = await _get_application_health(ctx, name)
    state = data["results"][0]["state"]

    if state == "RUNNING":
        await client.put(f"/applications/{name}/stop")

        while timeout:
            data = await _get_application_health(ctx, name)

            state = data["results"][0]["state"]

            if state == "STOPPED":
                break

            time.sleep(1)
            timeout -= 1

    elif state in ("DEAD", "DELETED"):
        raise errors.InvalidStateError(f"application `{name}` is `{state}`")

    if timeout == 0:
        raise errors.TimeoutExceededError()

    return {
        "name": name,
        "state": state
    }


async def restart_application(
    ctx: Annotated[Context, Field(
        description="The FastMCP Context object"
    )],
    name: Annotated[str, Field(
        description="The name of the application to restart"
    )],
    timeout: Annotated[int, Field(
        description="Timeout waiting for application to restart",
        default=10
    )]
) -> dict:
    """
    Restart an application on Itential Platform

    This tool will restart an application on an Itential Platform server.  The
    name argument defines the name of the application to restart.  In order to
    restart an application, the named application must be in a RUNNING state
    otherwise this tool will generate an error

    The name argument specifies the name of the appliciation to restart. If
    the specified applicatin does not exist, this tool will return a
    NotFoundError.  The name argument is case sensitive and must be a valid
    application name.  The list of applications can be found using the
    get_applications tool.

    The timeout value sets the number of seconds to wait when verifying the
    application has succussfully restarted.  If the timeout value expires
    before the application reaches a RUNNING state, a TimeoutExceededError
    will be generated.

    If the application is in a STOPPED, DEAD or DELETED state, an
    InvalidStateError is raised.

    Args:
        ctx (Context): The FastMCP Context object

        name (str): The name of the application to restart

        timeout (int): The timeout in seconds this tool should wait for the
            application to restart successfully.  If the application doesn't
            restart before the timeout expires, a TimeoutExceeded exception
            is raised

    Returns:
        dict: An object representing the restart operation
            - name: The name of the application
            - status: The state of the restart operation.  Valid values are
                RUNNING, DEAD, DELETED, STOPPED

    Raises:
        TimeoutExceededError: If the application isn't in a RUNNING state
            before the timeout expires

        InvalidStateError: if the application is in either the STOPPED, DEAD or
            DELETED state, this exception is raised

    Notes:
        - The application name argument is case senstive
        - The timeout value is set in seconds
        - If the application is in a STOPPED state, use the start_application
          tool instead

    """
    await ctx.info("inside restart_application(...)")

    client = ctx.request_context.lifespan_context.get("client")

    data = await _get_application_health(ctx, name)
    state = data["results"][0]["state"]

    if state == "RUNNING":
        await client.put(f"/applications/{name}/restart")

        while timeout:
            data = await _get_application_health(ctx, name)

            state = data["results"][0]["state"]

            if state == "RUNNING":
                break

            time.sleep(1)
            timeout -= 1

    elif state in ("DEAD", "DELETED", "STOPPED"):
        raise errors.InvalidStateError(f"application `{name}` is `{state}`")

    if timeout == 0:
        raise errors.TimeoutExceededError()

    return {
        "name": name,
        "state": state
    }



# Copyright (c) 2025 Itential, Inc
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

import pytest

from itential_mcp.stringutils import tostr, tobytes, toint, tobool


class TestStringUtils:

    @pytest.mark.parametrize("input_val,expected", [
        ("hello", "hello"),
        (123, "123"),
        (None, None),
        (True, "True"),
    ])
    def test_tostr(self, input_val, expected):
        assert tostr(input_val) == (str(expected) if expected is not None else expected)

    @pytest.mark.parametrize("input_val,encoding,expected", [
        ("hello", "utf-8", b"hello"),
        ("¡hola!", "utf-8", b"\xc2\xa1hola!"),
        ("test", "ascii", b"test"),
    ])
    def test_tobytes(self, input_val, encoding, expected):
        assert tobytes(input_val, encoding) == expected

    def test_tobytes_invalid_type(self):
        with pytest.raises(AttributeError):
            tobytes(None)

    @pytest.mark.parametrize("input_val,expected", [
        ("42", 42),
        ("0", 0),
        ("-15", -15),
    ])
    def test_toint_valid(self, input_val, expected):
        assert toint(input_val) == expected

    @pytest.mark.parametrize("input_val", ["abc", "", None])
    def test_toint_invalid(self, input_val):
        with pytest.raises((ValueError, TypeError)):
            toint(input_val)

    @pytest.mark.parametrize("input_val,expected", [
        ("true", True),
        ("True", True),
        ("1", True),
        ("yes", True),
        ("on", True),
        ("false", False),
        ("0", False),
        ("off", False),
        ("", False),
        (None, False),
    ])
    def test_tobool_string_inputs(self, input_val, expected):
        assert tobool(input_val) is expected

    def test_tobool_native_boolean(self):
        assert tobool(True) is True
        assert tobool(False) is False




# Copyright (c) 2025 Itential, Inc
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from typing import Annotated
from pydantic import Field

from fastmcp import Context

from itential_mcp import timeutils


async def _get_project_id_from_name(ctx: Context, name: str) -> str:
    """
    Gets the project ID for the specified project name

    This function will attempt to get the project id based on the name
    of the project.  It will query the server to get the project by name
    and return the project id.  If the project name doesn't not match
    on the server, a ValueError will be raised.

    Note the project name is case sensitive

    Args:
        ctx (Context): The FastMCP Context object
        name (str): The name of the project to locate

    Returns:
        str: The project ID associdated with the project name

    Raises:
        ValueError: Exception raised when the project name could not be
            deinitively located on the server
    """
    client = ctx.request_context.lifespan_context.get("client")

    res = await client.get(
        "/automation-studio/projects",
        params={"equals[name]": name}
    )

    data = res.json()

    if len(data["data"]) != 1:
        raise ValueError(f"unable to locate project `{name}`")

    return data["data"][0]["_id"]


async def get_command_templates(
    ctx: Annotated[Context, Field(
        description="The FastMCP Context object"
    )]
) -> list[dict]:
    """
    Get a list of command templates from Itential Platform

    This tool will get the list of command templates from Itential
    Platform.  It will retreive command templates defined in global
    space and in projects and return them as a list of Python dict
    elements.

    The elements contain the following fields:

        *_id: The unique identifier for this element
        * name: The name of the command template
        * description: Short description of the command template
        * namespace: Defines the project the template is part of.  If this
            field is null, the command template is in the global namespace
        * passRule: Configures the rules for passing.  When this value is
            set to True, all commands must pass and when this value is
            set to False, only one of the define commands must pass
        * created: ISO 8601 timestamp of when the trigger was created
        * createdBy: Account name that created the trigger
        * updated: ISO 8601 timestamp of when the trigger was last updated
        * updatedBy: Account name that last updated the trigger

    Args:
        ctx (Context): The FastMCP Context object

    Returns:
        list[dict]: A Python list of dict objects that represent the available
            workflows found on the server.

    Raises:
        None
    """
    await ctx.info("inside get_command_templates(...)")

    client = ctx.request_context.lifespan_context.get("client")

    res = await client.get("/mop/listTemplates")

    results = list()

    for item in res.json():
        results.append({
            "_id": item["_id"],
            "name": item["name"],
            "description": item["description"],
            "namespace": item["namespace"],
            "passRule": item["passRule"],
            "created": timeutils.epoch_to_timestamp(item["created"]),
            "createdBy": item["createdBy"],
            "updated": timeutils.epoch_to_timestamp(item["lastUpdated"]),
            "updatedBy": item["lastUpdatedBy"],
        })

    return results

async def describe_command_template(
    ctx: Annotated[Context, Field(
        description="The FastMCP Context object"
    )],
    name: Annotated[str, Field(
        description="The name of the command template to describe"
    )],
    project: Annotated[str | None, Field(
        description="The name of the project to get the command template from",
        default=None
    )]
) -> dict:
    """
    Get details about a specific command template

    This tool will retrieve a specific command template defined by the
    `name` argument and return its details.  If the command template is
    in a project, the project name must also be provided using the optional
    `project` argument.  If the `project` argument is None, only command
    templates defined in global space will be considered.

    The tool will return a dict object that represents the command template
    with the following structure:

        * _id: The unique identifier for this command template
        * name: The name of the command template
        * commands: The list of commands and rules associatedw this command
            template
        * namespace: Defines the project the template is part of.  If this
            field is null, the command template is in the global namespace
        * passRule: Configures the rules for passing.  When this value is
            set to True, all commands must pass and when this value is
            set to False, only one of the define commands must pass
        * created: ISO 8601 timestamp of when the trigger was created
        * createdBy: Account name that created the trigger
        * updated: ISO 8601 timestamp of when the trigger was last updated
        * updatedBy: Account name that last updated the trigger

    Args:
        ctx (Context): The FastMCP Context object
        name (str): The name of the command template to run
        devices (list): A list of devices to run the command template
            against.  The devices in this list must be known to Itential
            Platform.  To see the list of devices, use the `get_devices(...)`
            tool.

    Returns:
        list[dict]: A Python list of dict objects that represent the available
            workflows found on the server.

    Raises:
        None
    """
    await ctx.info("inside describe_command_template(...)")

    client = ctx.request_context.lifespan_context.get("client")

    if project is not None:
        project_id = await _get_project_id_from_name(ctx, project)
        name = f"@{project_id}: {name}"

    res = await client.get(f"/mop/listATemplate/{name}")

    data = res.json()[0]

    return {
        "_id": data["_id"],
        "name": data["name"],
        "passRule": data["passRule"],
        "commands": data["commands"],
        "created": timeutils.epoch_to_timestamp(data["created"]),
        "createdBy": data["createdBy"],
        "updated": timeutils.epoch_to_timestamp(data["lastUpdated"]),
        "updatedBy": data["lastUpdatedBy"],
        "namespace": data["namespace"]
    }


async def run_command_template(
    ctx: Annotated[Context, Field(
        description="The FastMCP Context object"
    )],
    name: Annotated[str, Field(
        description="The name of the command template to run"
    )],
    devices: Annotated[list, Field(
        description="The list of devices to run the command template against"
    )],
    project: Annotated[str | None, Field(
        description="Project that contains the command template",
        default=None
    )]
) -> dict:
    """
    Runs the command template against the list of devices

    This tool will run the named command template against one or more
    devices defined in the `devices` argument and return the results.  The
    response is an array that includes an element for each device response
    for each command run.

    Command Results
    The command results return the following keys:

        * raw: The command executed on the remote device
        * all_pass_flag: Boolean that indicates whether or not all rules
            must pass
        * evaluated: The command sent to the device
        * parameters: ???
        * rules: One or more rules to be evaluated when the command
            template is run.  See Rules for details on the returned
            object
        * device: The name of the device associated with this result
        * response: The response from the device used to run rules against
        * result: ???

    Rules:
    A command template can define one or more rules to validate the response
    against.  Rules are defined using the following structure:

        * eval: Type of rule evaluation to be performed
        * rule: The data to use for performing the rule check
        * severity: The severity of the error if the rule matches
        * raw: The raw data used when performing the rule check
        * result: The result from the rule check

    Once all commands have been executed against all devices and all rules
    have been processed, this function will return the final results

        * name: The name of the command template that was executed
        * all_pass_flag: Boolean that indicates whether or not all rules
            must pass
        * command_results: A list of elements one for each command executed
            on a device.  See `Command Results` for details about the return
            fields.

    Args:
        ctx (Context): The FastMCP Context object
        name (str): The name of the command template to run
        devices (list): A list of devices to run the command template
            against.  The devices in this list must be known to Itential
            Platform.  To see the list of devices, use the `get_devices(...)`
            tool.

    Returns:
        list[dict]: A Python list of dict objects that represent the available
            workflows found on the server.

    Raises:
        None
    """
    await ctx.info("inside run_command_templates(...)")

    client = ctx.request_context.lifespan_context.get("client")

    if project is not None:
        project_id = await _get_project_id_from_name(ctx, project)
        name = f"@{project_id}: {name}"

    body = {
        "template": name,
        "devices": devices,
    }

    res = await client.post("/mop/RunCommandTemplate", json=body)

    return res.json()


async def run_command(
    ctx: Annotated[Context, Field(
        description="The FastMCP Context object"
    )],
    cmd: Annotated[str, Field(
        description="The command to run on the devices"
    )],
    devices: Annotated[list[str], Field(
        description="The list of devices to run the command on"
    )]
) -> list[dict]:
    """
    Run a command against one or more devices

    This tool will run a command defined in the `cmd` argument against a
    list of devices from Itential Platform.  The command responses are
    returned.   The devices list must be the name of the device as known
    to Itential Platform.  To get a list of all devices using the
    `get_devices(...)` tool.

    The response is a list of dict elements with the following defined
    fields:

        * device: The name of the device the command was run against
        * command: The command sent to the device
        * response: The output from running the command on the device

    Args:
        ctx (Context): The FastMCP Context object
        name (str): The name of the command template to run
        devices (list): A list of devices to run the command template
            against.  The devices in this list must be known to Itential
            Platform.  To see the list of devices, use the `get_devices(...)`
            tool.

    Returns:
        list[dict]: A list of dict objects that represent the results from
            running the command on a device

    Raises:
        None
    """
    await ctx.info("inside run_command(...)")

    client = ctx.request_context.lifespan_context.get("client")

    body = {
        "command": cmd,
        "devices": devices,
    }

    res = await client.post("/mop/RunCommandDevices", json=body)

    results = list()

    for item in res.json():
        results.append({
            "device": item["device"],
            "command": item["raw"],
            "response": item["response"],
        })

    return results



# Copyright (c) 2025 Itential, Inc
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

import pytest
import httpx
from unittest.mock import AsyncMock, patch

from itential_mcp.client import PlatformClient
from itential_mcp.response import Response


@pytest.fixture
def mock_httpx_response():
    return httpx.Response(
        status_code=200,
        json={"result": "success"},
        request=httpx.Request("GET", "https://example.com")
    )


@pytest.fixture
def patched_platform_factory(mock_httpx_response):
    with patch("itential_mcp.client.ipsdk.platform_factory") as factory_mock:
        client_mock = AsyncMock()
        client_mock._send_request.return_value = mock_httpx_response
        factory_mock.return_value = client_mock
        yield factory_mock


@pytest.mark.asyncio
async def test_init_client(patched_platform_factory):
    client = PlatformClient()
    assert patched_platform_factory.called
    assert isinstance(client.client, AsyncMock)


@pytest.mark.asyncio
async def test_send_request(patched_platform_factory, mock_httpx_response):
    client = PlatformClient()
    response = await client.send_request("GET", "/test", params={"x": "1"}, json={"k": "v"})
    assert isinstance(response, Response)
    assert response.status_code == 200
    assert response.json() == {"result": "success"}


@pytest.mark.asyncio
async def test_get_method(patched_platform_factory, mock_httpx_response):
    client = PlatformClient()
    response = await client.get("/test", params={"a": "b"})
    assert isinstance(response, Response)
    assert response.status_code == 200


@pytest.mark.asyncio
async def test_post_method(patched_platform_factory, mock_httpx_response):
    client = PlatformClient()
    response = await client.post("/test", json={"data": 123})
    assert isinstance(response, Response)
    assert response.status_code == 200


@pytest.mark.asyncio
async def test_put_method(patched_platform_factory, mock_httpx_response):
    client = PlatformClient()
    response = await client.put("/test", json={"update": True})
    assert isinstance(response, Response)
    assert response.status_code == 200


@pytest.mark.asyncio
async def test_delete_method(patched_platform_factory, mock_httpx_response):
    client = PlatformClient()
    response = await client.delete("/test", params={"id": "1"})
    assert isinstance(response, Response)
    assert response.status_code == 200




# Copyright (c) 2025 Itential, Inc
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from typing import Annotated

from pydantic import Field

from fastmcp import Context


async def render_template(
    ctx: Annotated[Context, Field(
        description="The FastMCP Context object"
    )],
    template: Annotated[str, Field(
        description="The Jinaj2 template string"
    )],
    variables: Annotated[dict, Field(
        description="Zero or more variables to associate with this template",
        default=None
    )]
) -> str:
    """
    Renders a Jinja2 template

    This tool will take a Jinja2 template and set of variables and render
    the string.  The template argument defines the Jinja2 template and
    the variables argument is a Python dict that defines the values.

    The returned string is the fully rendered Jinaj2 template.

    Args:
        ctx (Context): The FastMCP Context object

        template (str): The Jinja2 template string to render

        variables (dict): Zero or more key value pairs as a Python dict
            to add to the configuration template

    Returns:
        str: A Python dict object that represents the rendered template

    Raises:
        None
    """
    await ctx.info("inside render_template()")

    client = ctx.request_context.lifespan_context.get("client")

    body = {
        "template": template,
        "variables": variables or {}
    }

    res = await client.post(
        "/configuration_manager/jinja2",
        json=body
    )

    return res.json()



# Copyright (c) 2025 Itential, Inc
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from typing import Annotated

from pydantic import Field

from fastmcp import Context

from itential_mcp import functions


async def get_resources(
    ctx: Annotated[Context, Field(
        description="The FastMCP Context object"
    )]
) -> list[dict]:
    """
    Get all Lifecycle Manager resource models from the server

    This tool will get all of the configured Lifecycle Manager resource models
    and return them as elements in a list.   The tool will return a list where
    each element represents a configured Lifecycle Model.

    The fields available for each element include:

        * _id: The unique identifier for this route
        * name: The name of the resource model
        * description: Short description of the model

    Args:
        ctx (Context): The FastMCP Context object

    Returns:
        list[dict]: A Python list of dict objects that represent the available
            resources found on the server.

    Raises:
        None
    """
    await ctx.info("inside get_resources(...)")

    client = ctx.request_context.lifespan_context.get("client")

    limit = 100
    skip = 0

    params = {"limit": limit}

    results = list()

    while True:
        params["skip"] = skip

        res = await client.get(
            "/lifecycle-manager/resources",
            params=params,
        )

        data = res.json()

        for item in data.get("data") or list():
            results.append({
                "_id": item["_id"],
                "name": item["name"],
                "description": item["description"],
            })

        if len(results) == data["metadata"]["total"]:
            break

        skip += limit

    return results


async def create_resource(
    ctx: Annotated[Context, Field(
        description="The FastMCP Context object"
    )],
    name: Annotated[str, Field(
        description="The name of the resource model to describe"
    )],
    schema: Annotated[dict, Field(
        description="JSON Schema representation of this resource"
    )],
    description: Annotated[str | None, Field(
        description="Short description of this resource",
        default=None
    )]
) -> dict:
    """
    Create a new Lifecycle Manager resource in Itential Platform

    This tool will create a new Lifecycle Manager resource model on the
    server. The resource model defines the structure and validation rules
    for resource instances.

    IMPORTANT: The schema parameter should contain ONLY the core schema
    definition (type, properties, required, etc.) WITHOUT JSON Schema
    metadata fields like $schema, title, description, or examples.

    Args:
        ctx (Context): The FastMCP Context object

        name (str):  The name of the resource to create. This should be a
            simple string identifier for the resource, e.g.,
            "PE-CE Network Connection" NOT a JSON object or schema document.

        schema (dict): The core schema definition object that defines the
            structure of the resource. This should include:
            - type: Usually "object"
            - properties: The properties definition
            - required: List of required property names

            DO NOT INCLUDE:
            - $schema: JSON Schema version reference
            - title: Schema title (use the 'name' parameter instead)
            - description: Schema description (use the 'description' parameter)
            - examples: Example values

            Example of CORRECT schema format:
            {
                "type": "object",
                "required": ["field1", "field2"],
                "properties": {
                    "field1": {
                        "type": "string",
                        "minLength": 1
                    },
                    "field2": {
                        "type": "number",
                        "minimum": 0
                    }
                }
            }

        description (str, optional): A human-readable description of what this
            resource represents. This is stored separately from the schema and
            used for documentation purposes.
            Example: "Provider Edge router with CE connections and billing info"

    Returns:
        dict: An object representing the created resource with fields:
            - _id: The unique identifier assigned by Itential
            - name: The resource name as provided
            - description: The description if provided
            - schema: The schema definition as stored
            - created: Timestamp of creation
            - createdBy: User who created the resource

    Raises:
        ValueError: Raised if:
            - The specified resource name already exists on Itential Platform
            - The schema format is invalid
            - Required parameters are missing or malformed

    Example Usage:
        # CORRECT way to create a resource:
        result = create_resource(
            ctx,
            name="Network Device",
            schema={
                "type": "object",
                "required": ["hostname", "ipAddress"],
                "properties": {
                    "hostname": {"type": "string"},
                    "ipAddress": {"type": "string", "format": "ipv4"}
                }
            },
            description="Basic network device configuration"
        )

        # WRONG - Don't pass the entire JSON Schema document as name:
        # result = create_resource(ctx, name=full_json_schema_doc, ...)

        # WRONG - Don't include metadata in schema:
        # result = create_resource(ctx, name="Device", schema={
        #     "$schema": "http://json-schema.org/draft-07/schema#",
        #     "title": "Device",  # Don't include this
        #     "type": "object",
        #     ...
        # })

        Notes:
            - The schema parameter defines the validation rules for resource instances
            - Once created, the resource can be used to create multiple instances
            - The schema follows JSON Schema draft-07 specification for validation
            - Metadata fields should be passed as separate parameters, not in the schema
    """
    await ctx.info("inside create_resource(...)")

    client = ctx.request_context.lifespan_context.get("client")

    existing = None
    try:
        existing = await describe_resource(ctx, name)
    except ValueError:
        pass

    if existing is not None:
        raise ValueError(f"resource {name} already exists")

    body = {
        "name": name,
        "schema": schema
    }

    if description is not None:
        body["description"] = description

    res = await client.post(
        "/lifecycle-manager/resources",
        json=body
    )

    data = res.json()["data"]

    return {
        "_id": data["_id"],
        "name": data["name"],
        "description": data["description"],
        "schema": data["schema"]
    }


async def describe_resource(
    ctx: Annotated[Context, Field(
        description="The FastMCP Context object"
    )],
    name: Annotated[str, Field(
        description="The name of the resource model to describe"
    )]
) -> dict:
    """
    Describe a Lifecycle Manager resource identified by name

    This tool will retrieve the Lifecycle Manager resource using the specified
    `name` argument and return it.   The returned value will be a Python
    dict object that represents the Lifecycle Manager resource.

    The returned object includes the following fields:

        * _id: The resource unique identifier
        * name: The name of the resource
        * description: Short description of the model
        * schema: The JSON schema that defines the resource which can be used
            to create a new instance of the resource
        * actions: The list of actions assoicated with this resource.  Actions
            can be invoked on instances of the resource to transition from
            one state to another

    Args:
        ctx (Context): The FastMCP Context object

        name (str): The name of the resource to get from the server

    Returns:
        dict: A Python dict object that represents the Lifecycle
            Manager resource

    Raises:
        ValueError: Raised if the specified Lifecycle Manager could not
            be uniquely identified on the server
    """
    await ctx.info("inside describe_resource(...)")

    client = ctx.request_context.lifespan_context.get("client")

    res = await client.get(
        "/lifecycle-manager/resources",
        params={"equals[name]": name},
    )

    data = res.json()

    if data["metadata"]["total"] != 1:
        raise ValueError(f"error attempting to find resource {name}")

    item = data["data"][0]

    actions = list()

    for ele in item["actions"]:
        if ele["workflow"] is not None:
            ele["workflow"] = await functions.workflow_id_to_name(ctx, ele["workflow"])

        if ele["preWorkflowJst"] is not None:
               ele["preWorkflowJst"] = await functions.transformation_id_to_name(ctx, ele["preWorkflowJst"])

        if ele["postWorkflowJst"] is not None:
               ele["postWorkflowJst"] = await functions.transformation_id_to_name(ctx, ele["postWorkflowJst"])

        actions.append(ele)

    return {
        "_id": item["_id"],
        "name": item["name"],
        "description": item["description"],
        "schema": item["schema"],
        "actions": actions,
    }


async def get_instances(
    ctx: Annotated[Context, Field(
        description="The FastMCP Context object"
    )],
    resource: Annotated[str, Field(
        description="The Lifecycle Manager resource name to retrieve instances for"
    )]
) -> list[dict]:
    """
    Get all instances for the resource from Itential Platform

    This tool will take the name of a Lifecycle Manager resource name and
    retrieve all configured instances of that resource.  The set of instances
    will be returned as a list.

    Each element in the list returns the following fields:

        * _id: The unique identifier for this route
        * name: The name of the resource model
        * description: Short description of the model
        * instanceData: Data object associated with this instance
        * lastAction: The last action performed on the instance

    Args:
        ctx (Context): The FastMCP Context object

        resource (str): The name of the resource to get instances for

    Returns:
        list[dict]: A list of Python dict objects where each element
            represents an instance of the resource

    Raises:
        ValueError: Raised if the specified resource could not be uniquely
            identified on the server

    """
    await ctx.info("inside get_instances(...)")

    client = ctx.request_context.lifespan_context.get("client")

    model_id = await functions.resource_name_to_id(ctx, resource)

    limit = 100
    skip = 0

    params = {"limit": limit}

    results = list()

    while True:
        params["skip"] = skip

        res = await client.get(
            f"/lifecycle-manager/resources/{model_id}/instances",
            params=params
        )

        data = res.json()

        for ele in data.get("data") or {}:
            results.append({
                "_id": ele["_id"],
                "name": ele["name"],
                "description": ele["description"],
                "instanceData": ele["instanceData"],
                "lastAction": ele["lastAction"],
            })

        if len(results) == data["metadata"]["total"]:
            break

        skip += limit

    return results



# Copyright (c) 2025 Itential, Inc
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

import os

from itential_mcp import cli

def test_parse_args_defaults(monkeypatch):
    monkeypatch.delenv("ITENTIAL_MCP_TRANSPORT", raising=False)
    cli.parse_args([])

    assert "ITENTIAL_MCP_SERVER_TRANSPORT" not in os.environ


def test_parse_args_custom_values():
    custom_args = [
        "--transport", "sse",
        "--host", "myhost",
        "--port", "9000",
        "--log-level", "DEBUG",
        "--platform-host", "platformhost",
        "--platform-user", "testuser",
        "--platform-password", "testpass"
    ]
    cli.parse_args(custom_args)

    assert os.environ["ITENTIAL_MCP_SERVER_TRANSPORT"] == "sse"
    assert os.environ["ITENTIAL_MCP_SERVER_HOST"] == "myhost"
    assert os.environ["ITENTIAL_MCP_SERVER_PORT"] == "9000"
    assert os.environ["ITENTIAL_MCP_SERVER_LOG_LEVEL"] == "DEBUG"
    assert os.environ["ITENTIAL_MCP_PLATFORM_HOST"] == "platformhost"
    assert os.environ["ITENTIAL_MCP_PLATFORM_USER"] == "testuser"


def test_env_variable_override(monkeypatch):
    monkeypatch.setenv("ITENTIAL_MCP_HOST", "envhost")
    monkeypatch.delenv("ITENTIAL_MCP_SERVER_HOST", raising=False)

    monkeypatch.setenv("ITENTIAL_MCP_PORT", "1234")
    monkeypatch.delenv("ITENTIAL_MCP_SERVER_PORT", raising=False)

    cli.parse_args([])

    assert os.environ["ITENTIAL_MCP_SERVER_HOST"] == "envhost"
    assert os.environ["ITENTIAL_MCP_SERVER_PORT"] == "1234"  # Still a string because it comes from env


def test_platform_env_variables_set(monkeypatch):
    monkeypatch.delenv("ITENTIAL_MCP_PLATFORM_USER", raising=False)
    cli.parse_args(["--platform-user", "envonly"])
    assert os.environ["ITENTIAL_MCP_PLATFORM_USER"] == "envonly"




# Copyright (c) 2025 Itential, Inc
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

import os
import sys
import asyncio
import argparse
import traceback

from collections.abc import Sequence
from dataclasses import fields

from . import server
from . import config


LEGACY_ENV_VARS = frozenset((
    ("ITENTIAL_MCP_TRANSPORT", "ITENTIAL_MCP_SERVER_TRANSPORT"),
    ("ITENTIAL_MCP_HOST", "ITENTIAL_MCP_SERVER_HOST"),
    ("ITENTIAL_MCP_PORT", "ITENTIAL_MCP_SERVER_PORT"),
    ("ITENTIAL_MCP_LOG_LEVEL", "ITENTIAL_MCP_SERVER_LOG_LEVEL"),
))


def parse_args(args: Sequence) -> None:
    """
    Parses any arguments

    This function will parse the arguments identified by the `args` argument
    and return a Namespace object with the values. Typically this is used
    to parse command line arguments passed when the application starts.

    Args:
        args (Sequence): The list of arguments to parse

    Returns:
        None

    Raises:
        None
    """
    parser = argparse.ArgumentParser(prog="itential-mcp")

    parser.add_argument(
        "--config",
        help="The Itential MCP configuration file"
    )

    # MCP Server arguments
    server_group = parser.add_argument_group(
        "MCP Server",
        "Configuration options for the MCP Server instance"
    )

    # Itential Platform arguments
    platform_group = parser.add_argument_group(
        "Itential Platform",
        "Configuration options for connecting to Itential Platform server"
    )

    data = [f for f in fields(config.Config)]

    for ele in data:
        attrs = ele.default.json_schema_extra
        if attrs and attrs.get("x-itential-mcp-cli-enabled"):
            helpstr = ele.default.description
            if helpstr is not None:
                helpstr += f" (default={ele.default.default})"
            else:
                helpstr = "NO HELP AVAILABLE!!"

            kwargs = {
                "dest": ele.name,
                "help": helpstr
            }

            kwargs.update(attrs.get("x-itential-mcp-options") or {})
            posargs = attrs.get("x-itential-mcp-arguments")


        if ele.name.startswith("server"):
            server_group.add_argument(*posargs, **kwargs)
        elif ele.name.startswith("platform"):
            platform_group.add_argument(*posargs, **kwargs)


    args = parser.parse_args(args=args)

    for key, value in dict(args._get_kwargs()).items():
        envkey = f"ITENTIAL_MCP_{key}".upper()
        if key.startswith("platform") or key.startswith("server"):
            if value is not None:
                if envkey not in os.environ:
                    if isinstance(value, str):
                        value = ", ".join(value.split(","))
                    os.environ[envkey] = str(value)

    conf_file = args.config
    if conf_file is not None:
        os.environ["ITENTIAL_MCP_CONFIG"] = conf_file

    # XXX (privateip) This will check for any values that use the legacy
    # environment variables which did not include the _SERVER_ in the name.
    for oldvar, newvar in LEGACY_ENV_VARS:
        if oldvar in os.environ and newvar not in os.environ:
            os.environ[newvar] = os.environ.pop(oldvar)


def run() -> int:
    """
    Main entry point for the application

    Args:
        None

    Returns:
        int:

    Raises:
        None
    """
    try:
        parse_args(sys.argv[1:])
        return asyncio.run(server.run())
    except Exception:
        traceback.print_exc()
        sys.exit(1)



# Copyright (c) 2025 Itential, Inc
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from importlib.metadata import version

from .cli import run

__version__ = version("itential_mcp")

__all__ = ("run",)



# Copyright (c) 2025 Itential, Inc
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from fastmcp import Context

async def workflow_id_to_name(ctx: Context, workflow_id: str) -> str:
    """
    Retrieves the workflow name for the specified workflow id

    This function will attempt to get the name of a workflow based on the
    specified workflow id.   The function will cache the workflow name
    to avoid uncessary lookups.

    Args:
        ctx (Context): The FastMCP Context object

        workflow_id (str): The workflow ID to find the name for

    Returns:
        str: The workflow name based on the specified workflow ID

    Raises:
        ValueError: If the specified workflow id cannot be found
    """
    cache = ctx.request_context.lifespan_context.get("cache")

    value = cache.get(f"/workflows/{workflow_id}")
    if value is not None:
        return value

    client = ctx.request_context.lifespan_context.get("client")

    res = await client.get(
        "/automation-studio/workflows",
        params={
            "equals[_id]": workflow_id,
            "include": "_id,name",
        }
    )

    data = res.json()

    if data["total"] == 0:
        raise ValueError(f"unable to find workflow with id {workflow_id}")

    value = data["items"][0]["name"]

    cache.put(f"/workflows/{workflow_id}", value)

    return value


async def account_id_to_username(ctx: Context, account_id: str) -> str:
    """
    Retrieves the username for an account id

    This function will take an account id and use it to look up the username
    associated with it.  The function will cache the username value to
    avoid making duplicate calls to the server.

    Args:
        ctx (Context): The FastMCP Context object

        acccount_id (str): The ID of the account to lookup and return the
            username for

    Returns:
        str: The username assoicated with the account id

    Raises:
        None
    """
    cache = ctx.request_context.lifespan_context.get("cache")

    value = cache.get(f"/accounts/{account_id}")
    if value is not None:
        return value

    client = ctx.request_context.lifespan_context.get("client")

    limit = 100
    skip = 0
    cnt = 0

    params = {"limit": limit}

    while True:
        params["skip"] = skip

        res = await client.get(
            "/authorization/accounts",
            params=params
        )

        data = res.json()
        results = data["results"]

        for item in results:
            if item["_id"] == account_id:
                value = item["username"]
                break

        cnt += len(results)

        if cnt == data["total"]:
            break

        skip += limit

    if value is None:
        raise ValueError(f"unable to find account with id {account_id}")

    cache.put(f"/accounts/{account_id}", value)

    return value


async def transformation_id_to_name(ctx: Context, jst_id: str) -> str:
    """
    Retrieves the transformation name for the specified transformation id

    This function will attempt to get the name of a transformation based on the
    specified transformation id.   The function will cache the transformation
    name to avoid uncessary lookups.

    Args:
        ctx (Context): The FastMCP Context object

        jst_id (str): The transformation ID to find the name for

    Returns:
        str: The transformation name based on the specified transformation ID

    Raises:
        None
    """
    cache = ctx.request_context.lifespan_context.get("cache")

    value = cache.get(f"/transformations/{jst_id}")
    if value is not None:
        return value

    client = ctx.request_context.lifespan_context.get("client")
    res = await client.get(f"/transformations/{jst_id}")

    data = res.json()
    value = data["name"]

    cache.put(f"/transformations/{jst_id}", value)

    return value


async def resource_name_to_id(ctx: Context, name: str) -> str:
    """
    Retrieves the resource id for the specified resource name

    This function will attempt to get the id of a resource based on the
    specified resource name.   The function will cache the resource id
    to avoid uncessary lookups.

    Args:
        ctx (Context): The FastMCP Context object

        name (str): The resource name to find the reosurce ID for

    Returns:
        str: The resource ID of the resource based on the resource name

    Raises:
        ValueError: If the specific resource name could not be found on
            the server
    """
    cache = ctx.request_context.lifespan_context.get("cache")

    value = cache.get(f"/resources/{name}")
    if value is not None:
        return value

    client = ctx.request_context.lifespan_context.get("client")
    res = await client.get(
        "/lifecycle-manager/resources",
        params={"equals[name]": name}
    )

    data = res.json()
    if data["metadata"]["total"] != 1:
        raise ValueError(f"error locating resouce {name}")

    value = data["data"][0]["_id"]

    cache.put(f"/resources/{name}", value)

    return value


async def group_id_to_name(ctx: Context, group_id: str) -> str:
    """
    Retrieves the group anme for the specified group id

    This function will attempt to get the group name for the specified
    group ID.

    Args:
        ctx (Context): The FastMCP Context object

        group_id (str): The group ID to find the group name for

    Returns:
        str: The name of the group associated with this group ID

    Raises:
    """
    cache = ctx.request_context.lifespan_context.get("cache")

    value = cache.get(f"/authorization/groups/{group_id}")
    if value is not None:
        return value

    client = ctx.request_context.lifespan_context.get("client")

    res = await client.get(f"/authorization/groups/{group_id}")

    data = res.json()

    value = data["name"]

    cache.put(f"/authorization/groups/{group_id}", value)


async def project_name_to_id(ctx: Context, name: str) -> str:
    """
    Retrieves the project ID for the specified project name

    This function will attempt to find the a project on the Itential
    Server with the name as specified by the name argument.   If the
    project exists, the project ID will be returned.  If the project
    does not exist, an exception will be raised

    Args:
        ctx (Context): The FastMCP Context object

        name (str): The project name to return the ID for

    Returns:
        str: The project ID of the project based on the name

    Raises:
        ValueError: If the specific project name could not be found on
            the server
    """
    cache = ctx.request_context.lifespan_context.get("cache")

    value = cache.get(f"/projects/{name}")
    if value is not None:
        return value

    client = ctx.request_context.lifespan_context.get("client")

    res = await client.get(
        "/automation-studio/projects",
        params={"equals[name]": name}
    )

    data = res.json()
    if data["metadata"]["total"] != 1:
        raise ValueError(f"error locating project {name}")

    value = data["data"][0]["_id"]

    cache.put(f"/projects/{name}", value)

    return value



# Copyright (c) 2025 Itential, Inc
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from itential_mcp.toolutils import tags

def test_tags_decorator_single():
    @tags("public")
    def my_func():
        return "hello"

    assert hasattr(my_func, "tags")
    assert my_func.tags == ["public"]

def test_tags_decorator_multiple():
    @tags("system", "admin", "beta")
    def another_func():
        return 42

    assert hasattr(another_func, "tags")
    assert set(another_func.tags) == {"system", "admin", "beta"}

def test_tags_does_not_modify_function_behavior():
    @tags("alpha")
    def simple_func(x):
        return x * 2

    assert simple_func(4) == 8
    assert simple_func.tags == ["alpha"]

def test_tags_empty():
    @tags()
    def no_tags_func():
        return "none"

    assert hasattr(no_tags_func, "tags")
    assert no_tags_func.tags == []



# Copyright (c) 2025 Itential, Inc
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)


def tostr(s: str | None) -> str:
    """
    Convert a string to a string

    Args:
        o (Any): The object to attempt to convert to to a string

    Returns:
        str: The string representation of the object

    Raises:
        None
    """
    if s is not None:
        s = str(s)
    return s


def tobytes(s: str | None, encoding: str = "utf-8") -> bytes:
    """
    Convert a string into bytes using the specified encoding.

    Args:
        s (str): The input string to convert.
        encoding (str): The character encoding to use (default is 'utf-8').

    Returns:
        bytes: The encoded byte representation of the string.

    Raises:
        None
    """
    return s.encode(encoding)


def toint(value: str | None) -> int:
    """
    Convert a string representation of an integer to an int type.

    Args:
        value (str): A string to convert

    Returns:
        int: The integer value

    Raises:
        None
    """
    return int(value)


def tobool(value: str | None) -> bool:
    """
    Convert a string representation of a boolean to a bool type.

    Args:
        value (Optional[str]): A string like "true", "false", "1", "0", etc.

    Returns:
        bool: The boolean value corresponding to the input.

    Raises:
        None
    """
    if value is None:
        return False
    if isinstance(value, bool):
        return value
    return value.strip().lower() in {"true", "1", "yes", "on"}



# Copyright (c) 2025 Itential, Inc
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

import time
import threading

from typing import Any


class Cache(object):
    """
    The Cache class implements a basic, thread safe implementation of a
    key/value store that can be used to cache values.  One instantiated,
    any value can be inserted into the store for later retrieval.
    """

    def __init__(self, cleanup_interval: int = 10):
        """
        Create a new instance of Cache

        This will create a new instance of Cache that can be used to store
        values based on a string key.  The values can later be retrieved.  It
        is the responsibility of the implementation to handle any data
        serialization for the value.

        This object supports a time-to-live value for all entries in the
        store.   It will start a background thread that is responsible
        for iterating over values in the store and expiring them.   The
        clean up interval can be specified using the `cleanup_interval`
        argument.

        Args:
            cleanup_interval (int): The internal specified in seconds to run
                the key expiration process.  The default is 10 seconds.

        Returns:
            Cache: An instance of Cache

        Raises:
            None
        """
        self._store = {}
        self._lock = threading.Lock()
        self._cleanup_interval = cleanup_interval
        self._stop_event = threading.Event()
        self._thread = threading.Thread(target=self._cleanup_expired_keys, daemon=True)
        self._thread.start()

    def put(self, key: str, value: Any, ttl: int | None = None):
        """
        Put a new key into the store

        This method will put a new key into the store.  The value can be
        any Python object.  It is the calling functions responsibity to
        handle any data serialization.

        Args:
            key (str): The key used to store and retreive the value
            value (Any): The value to store with the key
            ttl (int): Time-to-live for the key.  When the TTL expires the
                key is pruged from the store

        Returns:
            None

        Raises:
            None
        """
        expiry = time.time() + ttl if ttl else None
        with self._lock:
            self._store[key] = (value, expiry)

    def get(self, key: str) -> Any:
        """
        Get the value of `key` from the store

        This method will retreive the value from the store based on the
        specified `key` argument.   If the key does not exist in the
        store or has expired, this method will return None.

        Args:
            key (str): The key to retrieve the value for

        Returns:
            Any: The value associated with the key.  If the key doesn't exist
                None is returned

        Raises:
            None
        """
        with self._lock:
            item = self._store.get(key)
            if not item:
                return None

            value, expiry = item
            if expiry and time.time() > expiry:
                del self._store[key]
                return None

            return value

    def delete(self, key: str):
        """
        Delete a key from the store

        This method will delete a previously inserted key from the store.
        If the specified key does not exist in the store, this method
        simply performs a nooop

        Args:
            key (str): The key to delete from the store

        Returns:
            None

        Raises:
            None
        """
        with self._lock:
            return self._store.pop(key, None)

    def keys(self) -> list[str]:
        """
        Returns the list of keys from the store

        This method will return the list of keys that have been added to
        the store that have not expired.

        Args:
            None

        Returns:
            list[str]: A list that represents all keys in the store

        Raises:
            None
        """
        now = time.time()
        with self._lock:
            return [
                k for k, (v, expiry) in self._store.items()
                if not expiry or now <= expiry
            ]

    def clear(self):
        """
        Remove all entries from the store

        This method will delete all entries from the store regardless of
        the type of entry or TTL value.  After calling this method there
        will be no entries in the store  This is a destructive operation
        that cannot be undone once called.

        Args:
            None

        Returns:
            None

        Raises:
            None
        """
        with self._lock:
            self._store.clear()

    def _cleanup_expired_keys(self):
        """
        Internal method that handles cleaning up expired keys
        """
        while not self._stop_event.is_set():
            time.sleep(self._cleanup_interval)
            now = time.time()
            with self._lock:
                expired_keys = [k for k, (v, expiry) in self._store.items() if expiry and now > expiry]
                for k in expired_keys:
                    del self._store[k]

    def stop(self):
        """
        Stop the background cleanup thread

        This method will gracefully stop the store service.  It will
        connect to the background thread that handles content
        expiration and stops the thread from running.  It will also
        clear out all entries from the store

        Calling this method is optional and not required for shutting
        down the service.

        Args:
            None

        Returns:
            None

        Raises:
            None
        """
        self._stop_event.set()
        self._thread.join()
        self.clear()



# Copyright (c) 2025 Itential, Inc
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

import time

from typing import Annotated

from pydantic import Field

from fastmcp import Context

from itential_mcp import errors


async def _get_adapter_health(
    ctx: Context,
    name: str
) -> dict:
    """
    Get the health of an adapter

    This internal function will return the the health of the adapter
    specified by the name argument.  If the specified adapter does not
    exist, the function will raise a NotFoundError.   If the adapter
    does exist, it will be returned as a dict object.

    Args:
        ctx (Context): The FastMCP Context object

        name (str): The name of the adapter to get.  This argument is
            case senstive.  To get a list of all available adapters on
            the server see ```get_adapters```

    Returns:
        dict: An object the represents the adapter health

    Raises:
        NotFoundError: If the adapter specified by name cannot be found
            on the server, a NotFoundError is returned.
    """
    client = ctx.request_context.lifespan_context.get("client")

    res = await client.get(
        "/health/adapters",
        params={
            "equals": name,
            "equalsField": "id"
        }
    )

    data = res.json()

    if data["total"] != 1:
        raise errors.NotFoundError(f"unable to find adapter {name}")

    return data


async def get_adapters(
    ctx: Annotated[Context, Field(
        description="The FastMCP Context object"
    )]
) -> dict:
    """
    Get all adapters from Itential Platform

    This tool will retrieve all of the adapters configured on the instance
    of Itential Plaform and return them.   The returned list will include
    the adapter name, package, version, description and running state.

    Args:
        ctx (Context): The FastMCP Context object

    Returns:
        list[dict]: A list of objects where each object represents an
            adapter with the following fields:
            - name: The name of the adapter
            - package: The NodeJS package that comprises the adapter
            - version: The version of the adapter on the server
            - description: The adapter description
            - state: The current operational state of the appliacation.  Valid
                states are DEAD, STOPPED, RUNNING, DELETED

    Raises:
        None
    """
    await ctx.info("inside get_adapters(...)")

    client = ctx.request_context.lifespan_context.get("client")

    res = await client.get("/health/adapters")

    data = res.json()

    results = list()

    for ele in data["results"]:
        results.append({
            "name": ele["id"],
            "package": ele.get("package_id"),
            "version": ele["version"],
            "description": ele.get("description"),
            "state": ele["state"],
        })

    return results


async def start_adapter(
    ctx: Annotated[Context, Field(
        description="The FastMCP Context object"
    )],
    name: Annotated[str, Field(
        description="The name of the adapter to start"
    )],
    timeout: Annotated[int, Field(
        description="Timeout waiting for adapter to start",
        default=10
    )]
) -> dict:
    """
    Start an adapter on Itential Platform

    This tool will start an adapter on an Itential Platform server.  The
    name argument defines the name of the adapter to start.  If the
    adapter is already in a running state. this tool will not attempt
    any further action.

    The name argument is case sensitive and must be a valid adapter
    defined on Itential Platform.  The list of adapters running on
    the server can be found using the get_adapters tool.

    If the adapter is in a STOPPED state, this tool will attempt to start
    the adapter.   It will check to see if the adapter started.  If
    the adapter hasn't started within the timeout period, a
    TimeoutExceededError error is raised

    If the adapter is in either the DEAD or DELETED state, an
    InvalidStateError is raised.

    Args:
        ctx (Context): The FastMCP Context object

        name (str): The name of the adapter to start

        timeout (int): The timeout in seconds this tool should wait for the
            adapter to start successfully.  If the adapter doesn't
            start before the timeout expires, a TimeoutExceeded exception
            is raised

    Returns:
        dict: An object representing the start operation
            - name: The name of the adapter
            - status: The state of the start operation.  Valid values are
                RUNNING, DEAD, DELETED, STOPPED

    Raises:
        TimeoutExceededError: If the adapter isn't in a RUNNING state
            before the timeout expires

        InvalidStateError: if the adapter is in either the DEAD or
            DELETED state, this exception is raised

    Notes:
        - The adapter name argument is case senstive
        - The timeout value is set in seconds

    """
    await ctx.info("inside start_adapter(...)")

    client = ctx.request_context.lifespan_context.get("client")

    data = await _get_adapter_health(ctx, name)
    state = data["results"][0]["state"]

    if state == "STOPPED":
        await client.put(f"/adapters/{name}/start")

        while timeout:
            data = await _get_adapter_health(ctx, name)
            state = data["results"][0]["state"]

            if state == "RUNNING":
                break

            time.sleep(1)
            timeout -= 1

    elif state in ("DEAD", "DELETED"):
        raise errors.InvalidStateError(f"adapter `{name}` is `{state}`")

    if timeout == 0:
        raise errors.TimeoutExceededError()

    return {
        "name": name,
        "state": state
    }


async def stop_adapter(
    ctx: Annotated[Context, Field(
        description="The FastMCP Context object"
    )],
    name: Annotated[str, Field(
        description="The name of the adapter to stop"
    )],
    timeout: Annotated[int, Field(
        description="Timeout waiting for adapter to start",
        default=10
    )]
) -> dict:
    """
    Stop an adapter on Itential Platform

    This tool will stop an adapter that is currently in a running
    state on the server.  If the adapter is already stopped this
    tool will not perform any further action.

    The name argument specifies the name of the adapter to stop.  The
    value for name must be a valid case senstive name.  The list of
    adapters can be found by calling the get_adapters tool.  If
    the specified name is not a valid adapter name, this tool will
    raise a NotFoundError.

    The timeout value sets the number of seconds to wait to verify the
    adapter has stopped.  If the timeout value expires before the
    adapter reaches a stopped state, this tool will raise a
    TimeoutExceededError.

    If the adapter is in either the DEAD or DELETED state, an
    InvalidStateError is raised.

    Args:
        ctx (Context): The FastMCP Context object

        name (str): The name of the adapter to start

        timeout (int): The timeout in seconds this tool should wait for the
            adapter to stop successfully.  If the adapter doesn't
            stop before the timeout expires, a TimeoutExceeded exception
            is raised

    Returns:
        dict: An object representing the stop operation
            - name: The name of the adapter
            - status: The state of the start operation.  Valid values are
                RUNNING, DEAD, DELETED, STOPPED

    Raises:
        TimeoutExceededError: If the adapter isn't in a STOPPED state
            before the timeout expires

        InvalidStateError: if the adapter is in either the DEAD or
            DELETED state, this exception is raised

    Notes:
        - The adapter name argument is case senstive
        - The timeout value is set in seconds
    """
    await ctx.info("inside stop_adapter(...)")

    client = ctx.request_context.lifespan_context.get("client")

    data = await _get_adapter_health(ctx, name)
    state = data["results"][0]["state"]

    if state == "RUNNING":
        await client.put(f"/adapters/{name}/stop")

        while timeout:
            data = await _get_adapter_health(ctx, name)

            state = data["results"][0]["state"]

            if state == "STOPPED":
                break

            time.sleep(1)
            timeout -= 1

    elif state in ("DEAD", "DELETED"):
        raise errors.InvalidStateError(f"adapter `{name}` is `{state}`")

    if timeout == 0:
        raise errors.TimeoutExceededError()

    return {
        "name": name,
        "state": state
    }


async def restart_adapter(
    ctx: Annotated[Context, Field(
        description="The FastMCP Context object"
    )],
    name: Annotated[str, Field(
        description="The name of the adapter to restart"
    )],
    timeout: Annotated[int, Field(
        description="Timeout waiting for adapter to restart",
        default=10
    )]
) -> dict:
    """
    Restart an adapter on Itential Platform

    This tool will restart an adapter on an Itential Platform server.  The
    name argument defines the name of the adapter to restart.  In order to
    restart an adapter, the named adapter must be in a RUNNING state
    otherwise this tool will generate an error

    The name argument specifies the name of the appliciation to restart. If
    the specified applicatin does not exist, this tool will return a
    NotFoundError.  The name argument is case sensitive and must be a valid
    adapter name.  The list of adapters can be found using the
    get_adapters tool.

    The timeout value sets the number of seconds to wait when verifying the
    adapter has succussfully restarted.  If the timeout value expires
    before the adapter reaches a RUNNING state, a TimeoutExceededError
    will be generated.

    If the adapter is in a STOPPED, DEAD or DELETED state, an
    InvalidStateError is raised.

    Args:
        ctx (Context): The FastMCP Context object

        name (str): The name of the adapter to restart

        timeout (int): The timeout in seconds this tool should wait for the
            adapter to restart successfully.  If the adapter doesn't
            restart before the timeout expires, a TimeoutExceeded exception
            is raised

    Returns:
        dict: An object representing the restart operation
            - name: The name of the adapter
            - status: The state of the restart operation.  Valid values are
                RUNNING, DEAD, DELETED, STOPPED

    Raises:
        TimeoutExceededError: If the adapter isn't in a RUNNING state
            before the timeout expires

        InvalidStateError: if the adapter is in either the STOPPED, DEAD or
            DELETED state, this exception is raised

    Notes:
        - The adapter name argument is case senstive
        - The timeout value is set in seconds
        - If the adapter is in a STOPPED state, use the start_adapter
          tool instead

    """
    await ctx.info("inside restart_adapter(...)")

    client = ctx.request_context.lifespan_context.get("client")

    data = await _get_adapter_health(ctx, name)
    state = data["results"][0]["state"]

    if state == "RUNNING":
        await client.put(f"/adapters/{name}/restart")

        while timeout:
            data = await _get_adapter_health(ctx, name)

            state = data["results"][0]["state"]

            if state == "RUNNING":
                break

            time.sleep(1)
            timeout -= 1

    elif state in ("DEAD", "DELETED", "STOPPED"):
        raise errors.InvalidStateError(f"adapter `{name}` is `{state}`")

    if timeout == 0:
        raise errors.TimeoutExceededError()

    return {
        "name": name,
        "state": state
    }



# Copyright (c) 2025 Itential, Inc
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from typing import Annotated

from pydantic import Field

from fastmcp import Context


async def get_compliance_plans(
    ctx: Annotated[Context, Field(
        description="The FastMCP Context object"
    )],
) -> list[dict]:
    """
    Gets all configured compliance plans from Itential Platform

    This tool will return a all of the configured compliance plans found
    on from Itential Platform.  Each element in the returned list represents
    a configured compliance plan.  If there are no configured compliance
    plans, the tool will return an empty list.

    Elements in the list have the following fields:

        * id: The unique identifier for the compliance plan
        * name: The name of the compliance plan
        * description: Short description about the compliance plan
        * throttle: The configured number of devices that will be checked
            in parallel when the compliance plan is run

    Args:
        ctx (Context): The FastMCP Context object

    Returns:
        dict: A Python list of dict objects that reprsesent all of the devices
            knownn to Itential Platform

    Raises:
        None
    """
    await ctx.info("inside get_compliance_plans(...)")

    client = ctx.request_context.lifespan_context.get("client")

    limit = 100
    start = 0

    results = list()

    while True:
        body = {
            "name": "",
            "options": {
                "start": start,
                "limit": limit
            }
        }

        res = await client.post(
            "/configuration_manager/search/compliance_plans",
            json=body,
        )

        data = res.json()

        print(data)

        for ele in data["plans"]:
            results.append({
                "id": ele["id"],
                "name": ele["name"],
                "description": ele["description"],
                "throttle": ele["throttle"]
            })

        if len(results) == data["totalCount"]:
            break

        start += limit

    return results


async def run_compliance_plan(
    ctx: Annotated[Context, Field(
        description="The FastMCP Context object"
    )],
    name: Annotated[str, Field(
        description="The name of the compliance plan to run"
    )]
) -> dict:
    """
    Run a compliance plan from Itential Platform

    This tool will start the running of the compliance plan specified
    by the name argument.  Once started, this tool will return the
    running instance of the compliance plan.  If the compliance plan
    does not exist, this tool will return an error.

    The name argument is case sensitive

    The returned object has the following fields:

        * id: The unique identifier for this compliance plan instance
        * name: The name of the compliance plan that was started
        * description: Short description of the compliance plan
        * jobStatus: The current job status of the compliance plan instance

    Args:
        ctx (Context): The FastMCP Context object

        name (str): The name of the compliance plan to run

    Returns:
        dict: an object that represents the running instance of the
            compliance plan

    Raises:
        ValueError: Raises when the compliance plan specified by the
            name argument is not found
    """
    await ctx.info("inside run_compliance_plan(...)")

    client = ctx.request_context.lifespan_context.get("client")

    plans = await get_compliance_plans(ctx)

    plan_id = None

    for ele in plans:
        if ele["name"] == name:
            plan_id = ele["id"]
            break
    else:
        raise ValueError(f"compliance plan {name} not found")

    await client.post(
        "/configuration_manager/compliance_plans/run",
        json={"planId": plan_id}
    )

    body = {
        "searchParams": {
            "limit": 1,
            "planId": plan_id,
            "sort": { "started": -1 },
            "start": 0
        }
    }

    res = await client.post(
        "/configuration_manager/search/compliance_plan_instances",
        json=body
    )

    instance = res.json()
    data = instance["plans"][0]

    return {
        "id": data["id"],
        "name": data["name"],
        "description": data["description"],
        "jobStatus": data["jobStatus"]
    }



# Copyright (c) 2025 Itential, Inc
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from typing import Annotated

from pydantic import Field

from fastmcp import Context


async def describe_compliance_report(
    ctx: Annotated[Context, Field(
        description="The FastMCP Context object"
    )],
    report_id: Annotated[str, Field(
        description="The ID of the report to describe"
    )]
) -> dict:
    """
    Retrieve the compliance report from the server

    Args:
        ctx (Context): The FastMCP Context object

        report_id (str): The unique report identifier

    Returns:
        dict: A Python dict object that represents the output from running a
            compliance report (golden config) against devices in the infrastructure

    Raises:
        None
    """
    await ctx.info("inside describe_compliance_report(...)")

    client = ctx.request_context.lifespan_context.get("client")

    res = await client.get(
        f"/configuration_manager/compliance_reports/details/{report_id}"
    )

    return res.json()



# Copyright (c) 2025 Itential, Inc
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from typing import Any

import httpx


class Response(object):

    def __init__(self, res: httpx.Response):
        """
        Initialize a new response instance

        Args:
            res: An instance of httpx.Response
        """
        self.response = res

    @property
    def status_code(self) -> int:
        """
        Returns the HTTP status code
        """
        return self.response.status_code

    @property
    def reason(self) -> str:
        """
        Returns the HTTP status code message
        """
        return self.response.reason_phrase

    @property
    def text(self) -> str:
        """
        Return the response body as text
        """
        return self.response.text

    def json(self) -> Any:
        """
        Attempt to return the response body as a Python object

        Args:
            None

        Returns:
            Any: Returns the response text as a object

        Raises:
            None
        """
        return self.response.json()



# Copyright (c) 2025 Itential, Inc
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

import os
import sys
import inspect
import importlib.util

from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager, AsyncExitStack

from typing import Any

from fastmcp import FastMCP

from . import client
from . import config
from . import cache


@asynccontextmanager
async def lifespan(mcp: FastMCP) -> AsyncGenerator[dict[str | Any], None]:
    """
    Manage the lifespan of Itential Platform servers

    This function is responsible for creating the client connection to
    Itential Platform and yielding it to FastMCP to be included in the
    context.

    Args:
        mcp (FastMCP): An instance of FastMCP

    Returns:
        AsyncGenerator: Yields an AsyncGenerator with a dict object

    Raises:
        None
    """
    async with AsyncExitStack():
        yield {
            "client": client.PlatformClient(),
            "cache": cache.Cache()
        }


def new(cfg: config.Config) -> FastMCP:
    """
    Initialize the FastMCP server

    This function will intialize a new instance of the FastMCP server and
    return it to the calling function.  This function should only be called
    once to initialize the server.

    Args:
        cfg (Config): An instance of `config.Config` that provides the
            server configuration values

    Returns:
        FastMCP: An instance of a FastMCP server

    Raises:
        None
    """
    # Initialize FastMCP server
    return FastMCP(
        name="Itential Platform MCP",
        instructions="Itential tools and resources for interacting with Itential Platform",
        lifespan=lifespan,
        include_tags=cfg.server.get("include_tags"),
        exclude_tags=cfg.server.get("exclude_tags")
    )


def register_tools(mcp: FastMCP) -> None:
    """
    Register all functions in the tools folder with mcp

    This function will recursively load all modules found in the `tools`
    folder as long as they module name does not start with underscore (_).  It
    will then inspect the module to find all public functions and attach
    them to the instance of mcp as a tool.

    Args:
        mcp (FastMCP): An instance of FastMCP to attach tools to

    Returns:
        None

    Raises:
        None
    """
    path = os.path.join(os.path.dirname(os.path.realpath(__file__)), "tools")

    # Get a list of all files in the directory
    module_files = [f[:-3] for f in os.listdir(path) if f.endswith(".py") and f != "__init__.py"]

    # Import the modules, add them to globals and mcp
    for module_name in module_files:
        if not module_name.startswith("_"):
            spec = importlib.util.spec_from_file_location(module_name, os.path.join(path, f"{module_name}.py"))
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)

            # Inspect the module to retreive all of the functions.
            for name, f in inspect.getmembers(module, inspect.isfunction):
                if not name.startswith("_") and f.__module__ == module_name:
                    kwargs = {"tags": {name}}
                    if hasattr(f, "tags"):
                        tags = kwargs["tags"]
                        for ele in f.tags:
                            tags.add(ele)
                        kwargs["tags"] = tags
                    mcp.tool(f, **kwargs)


async def run() -> int:
    """
    Run the MCP server

    This is the server entry point for running the Itential MCP server using
    either stdio or sse.  This function will load the configuration, create
    the MCP server, register all tools and start the server.

    Args:
        None

    Returns:
        int: Returns a int value as the return code from running the server
            A value of 0 is success and any other value is an error

    Raises:
        KeyboardInterrupt: When an operator uses a keyboard interrupt,
            typically CTRL-C to stop the server.  This will cause the
            server to exit with return code 0
        Exception: Generic exception caught while running the server and
            prints the traceback to stdout.  This will cause the server
            to exit with return code 1
    """
    cfg = config.get()

    mcp = new(cfg)

    try:
        register_tools(mcp)
    except Exception as exc:
        print(f"ERROR: failed to import tool: {str(exc)}", file=sys.stderr)
        sys.exit(1)

    kwargs = {
        "transport": cfg.server.get("transport")
    }

    if kwargs["transport"] in ("sse", "streamable-http"):
        kwargs.update({
            "host": cfg.server.get("host"),
            "port": cfg.server.get("port"),
            "log_level": cfg.server.get("log_level")
        })
        if kwargs["transport"] == "streamable-http":
            kwargs["path"] = cfg.server.get("path")

    try:
        await mcp.run_async(**kwargs)
    except KeyboardInterrupt:
        print("Shutting down the server")
        sys.exit(0)
    except Exception as exc:
        print(f"ERROR: server stopped unexpectedly: {str(exc)}", file=sys.stderr)
        sys.exit(1)



# Copyright (c) 2025 Itential, Inc
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from typing import Annotated

from pydantic import Field

from fastmcp import Context


async def get_devices(
    ctx: Annotated[Context, Field(
        description="The FastMCP Context object"
    )],
) -> list[dict]:
    """
    Retrieve all devies known to Itential Platform

    Itential Platform will federate device information from multiple
    sources and make it available to workflows for performing tasks
    against physical devices. This function will query Itential Platform
    and return all of the devices known to it.

    Args:
        ctx (Context): The FastMCP Context object

    Returns:
        dict: A Python list of dict objects that reprsesent all of the devices
            knownn to Itential Platform

    Raises:
        None
    """
    await ctx.info("inside get_devices(...)")

    client = ctx.request_context.lifespan_context.get("client")

    limit = 100
    start = 0

    results = list()

    while True:
        body = {
            "options": {
                "order": "ascending",
                "sort": [{"name": 1}],
                "start": start,
                "limit": limit
            }
        }

        res = await client.post(
            "/configuration_manager/devices",
            json=body,
        )

        data = res.json()

        results.extend(data["list"])

        if len(results) == data["return_count"]:
            break

        start += limit

    return results


async def get_device_configuration(
    ctx: Annotated[Context, Field(
        description="The FastMCP Context object"
        )],
    name: Annotated[str, Field(
        description="The name of the device to retrieve the configuration from"
    )]
) -> str:
    """
    Get the device current configuration

    This tool will get the current device configuration using Itential
    Platform.  The device configuration will be returned if the device
    is a valid device otherwise it will raise an error.

    Args:
        ctx (Context): The FastMCP Context object

        name (str): The name of the device to backup.  The device must be
            available in the list of devices returned from Itential Platform

    Returns:
        str: The current device configuration

    Raises:
        ValueError: When there is an exception making the API call
    """
    await ctx.info("inside get_device_configuration(...)")

    client = ctx.request_context.lifespan_context.get("client")

    try:
        res = await client.get(f"/configuration_manager/devices/{name}/configuration")
    except ValueError:
        raise

    return res.json()["config"]


async def backup_device_configuration(
    ctx: Annotated[Context, Field(
        description="The FastMCP Context object"
        )],
    name: Annotated[str, Field(
        description="The name of the device to backup"
    )],
    description: Annotated[str, Field(
        description="Short description to attach to the backup",
        default=None
    )],
    notes: Annotated[str, Field(
        description="Notes to attach to the backup",
        default=None
    )]
) -> dict:
    """
    Backup a device configuration to Itential Platform

    This tool will invoke an operation to backup the configuration of a
    specific device to Itential Platform.  The deivce's configuration will be
    backed up and added to the Itential Platform server.

    The returned object includes the following fields:

        * id: The unique identifer for the backup on the server.  This value
            can be used to retreive the backup later
        * status: The status of the backup operation job
        * message: A short descriptive message about the status of the job

    Args:
        ctx (Context): The FastMCP Context object

        name (str): The name of the device to backup.  The device must be
            available in the list of devices returned from Itential Platform

        description (str): Optional short description of the backup

        notes (str): Optional text notes to append to the backup for later
            reference

    Returns:
        dict: A Python dic object that returns the success or failure of the
            backup operation

    Raises:
        None
    """
    await ctx.info("inside backup_device_configuration(...)")

    client = ctx.request_context.lifespan_context.get("client")

    body = {
        "name": name,
        "options": {
            "description": description or "",
            "notes": notes or ""
        }
    }

    res = await client.post(
        "/configuration_manager/devices/backups",
        json=body
    )

    return res.json()


async def apply_device_configuration(
    ctx: Annotated[Context, Field(
        description="The FastMCP Context object"
        )],
    device: Annotated[str, Field(
        description="The name of the device to apply the configuation to"
    )],
    config: Annotated[str, Field(
        description="The configuration to apply to the device"
    )]
) -> dict:
    """
    Apply a configuration to a device in Itential Platform

    This tool will apply a  configuration to a named device using Itential
    Platform.  The name argument defines the name of the device to apply
    the configuration to.  The device name needs to be a valid device name
    as returned by the get_devices tool.   The config argument specifies
    the configuation string to send to the device.  The configuration
    argument needs to be string.

    Args:
        ctx (Context): The FastMCP Context object

        name (str): The name of the device to apply the configuration to

        config (str): The actual configuration to send to the device

    Returns:
        dict: An object that provides the results of the operation

    Raises:
        None
    """
    await ctx.info("inside apply_device_configuration(...)")

    client = ctx.request_context.lifespan_context.get("client")

    body = {
        "config": {
            "device": device,
            "config": config,
        }
    }

    res = await client.post(
        f"/configuration_manager/devices/{device}/configuration",
        json=body
    )

    return res.json()



# Copyright (c) 2025 Itential, Inc
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from typing import Annotated

from pydantic import Field

from fastmcp import Context


async def get_job_metrics(
    ctx: Annotated[Context, Field(
        description="The FastMCP Context object"
    )]
) -> list[dict]:
    """
    Get the aggregate job metrics from workflow engine

    The Itential Platform workflow engine maintains records that aggregate
    workflow job metrics over the life of automation executions. This
    function will return the aggregate job metrics for automations that
    have been executed by Itential Platform.

    It will provide information such as the number of jobs completed in the
    `jobsComplete` key, the name of the workflow that was executed and the
    total run time among other key statistics

    Args:
        ctx (Context): The FastMCP Context object

    Returns:
        list[dict]: A Python list object that contains job metric data as
            a Python dict object

    Raises:
        None
    """
    await ctx.debug("inside get_job_metrics(...)")

    client = ctx.request_context.lifespan_context.get("client")

    limit = 100
    skip = 0

    params = {"limit": limit}

    results = list()

    while True:
        params["skip"] = skip

        res = await client.get(
            "/workflow_engine/jobs/metrics",
            params=params,
        )

        data = res.json()
        elements = data.get("results") or list()

        results.extend(elements)

        if len(elements) == data["total"]:
            break

        skip += limit

    return results


async def get_task_metrics(
    ctx: Annotated[Context, Field(
        description="The FastMCP Context object"
    )]
) -> list[dict]:
    """
    Get the aggregate task metrics from workflow engine

    The Itential Platform workflow engine maintains records that aggregate
    workflow task metrics over the life of automation executions. This
    function will return the aggregate task metrics for automations that
    have been executed by Itential Platform.

    It will provide details about tasks from workflows such as the application
    associated with the task, the task name and type and metrics for the task.

    Args:
        ctx (Context): The FastMCP Context object

    Returns:
        list[dict]: A Python list object that contains task metric data as
            a Python dict object

    Raises:
        None
    """
    await ctx.debug("inside get_task_metrics(...)")

    client = ctx.request_context.lifespan_context.get("client")

    limit = 100
    skip = 0

    results = list()

    while True:
        res = await client.get("/workflow_engine/tasks/metrics")

        data = res.json()
        results.extend(data["results"])

        if len(results) == data["total"]:
            break

        skip += limit

    return results




# Copyright (c) 2025 Itential, Inc
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

import ipsdk

from ipsdk.platform import AsyncPlatform

import httpx

from . import config
from . import response


class PlatformClient(object):

    def __init__(self):
        self.client = self._init_client()

    def _init_client(self) -> AsyncPlatform:
        """
        Initializes the client connection to Itential Platform

        Args:
            None

        Returns:
            AsyncPlatform: An instance of AsyncPlatform

        Raises:
            None
        """
        cfg = config.get()
        return ipsdk.platform_factory(want_async=True, **cfg.platform)

    async def _make_response(self, res: httpx.Response) -> response.Response:
        """
        Creates a response object and returns it

        Args:
            res (httpx.Response): The response object returned from the HTTP API
                request

        Returns:
            Response: A HTTP Response object

        Raises:
            None
        """
        return response.Response(res)

    async def send_request(
        self,
        method: str,
        path: str,
        params: dict = None,
        json: str | bytes | dict | list | None = None
    ) -> response.Response:
        """
        Send the request to the server and return the response

        Args:
            method (str): The HTTP method to invoke. This should be one of
                "GET", "POST", "PUT", "DELETE"

            path (str): The full URL path to send the reques to

            params (dict): A Python dict objec to be converted into a query
                string and appeneded to the URL

            json (str|bytes|dict|list): A Python object that can be serialized
                into a JSON object.

        Returns:
            Response: The HTTP response from the server

        Raises:
            None
        """
        res = await self.client._send_request(method, path, params, json)
        return await self._make_response(res)

    async def get(
        self,
        path: str,
        params: dict | None = None
    ) -> response.Response:
        """
        Send a HTTP GET request to the server

        Args:
            path (str): The full path to send the HTTP request to

            params (dict): A Python dict object to be converted to a query
                string and appended to the path

        Returns:
            Response: An HTTP Response object from the server

        Raises:
            None
        """
        return await self.send_request(
            method="GET", path=path, params=params
        )

    async def post(
        self,
        path: str,
        params: dict | None = None,
        json: str | dict | list | None = None,
    ) -> response.Response:
        """
        Send a HTTP POST request to the server

        Args:
            path (str): The full path to send the HTTP request to

            params (dict): A Python dict object to be converted to a query
                string and appended to the path

            json (str | dict | list): A Python object that can be serialized
                to a JSON string and sent as the body of the request

        Returns:
            Response: An HTTP Response object from the server

        Raises:
            None
        """
        return await self.send_request(
            method="POST", path=path, params=params, json=json
        )

    async def put(
        self,
        path: str,
        params: dict | None = None,
        json: str | dict | list | None = None,
    ) -> response.Response:
        """
        Send a HTTP PUT request to the server
        Args:
            path (str): The full path to send the HTTP request to

            params (dict): A Python dict object to be converted to a query
                string and appended to the path

            json (str | dict | list): A Python object that can be serialized
                to a JSON string and sent as the body of the request
        Returns:
            Response: An HTTP Response object from the server

        Raises:
            None
        """
        return await self.send_request(
            method="PUT", path=path, params=params, json=json
        )

    async def delete(
        self,
        path: str,
        params: dict | None = None,
    ) -> response.Response:
        """
        Send a HTTP DELETE request to the server
        Args:
            path (str): The full path to send the HTTP request to

            params (dict): A Python dict object to be converted to a query
                string and appended to the path

            json (str | dict | list): A Python object that can be serialized
                to a JSON string and sent as the body of the request
        Returns:
            Response: An HTTP Response object from the server

        Raises:
            None
        """
        return await self.send_request(
            method="DELETE", path=path, params=params
        )
